# -*- coding: utf-8 -*-
import ctypes

uint8_t = ctypes.c_ubyte
char = ctypes.c_char
uint32_t = ctypes.c_uint
uint64_t = ctypes.c_uint64
uint16_t = ctypes.c_ushort
guid_t = char * 16

FIRMWARE_VOLUME_GUIDS = {
    "FFS1":        "7a9354d9-0468-444a-81ce-0bf617d890df",
    "FFS2":        "8c8ce578-8a3d-4f1c-9935-896185c32dd3",
    "FFS3":        "5473c07a-3dcb-4dca-bd6f-1e9689e7349a",
    "NVRAM_EVSA":  "fff12b8d-7696-4c8b-a985-2747075b4f50",
    "NVRAM_NVAR":  "cef5b9a3-476d-497f-9fdc-e98143e0422c",
    "NVRAM_EVSA2": "00504624-8a59-4eeb-bd0f-6b36e96128e0",
    "APPLE_BOOT":  "04adeead-61ff-4d31-b6ba-64f8bf901f5a",
    #'08758b38-458d-50e8-56e8-3bffffff83c4'
}

FIRMWARE_CAPSULE_GUIDS = [
    '3b6686bd-0d76-4030-b70e-b5519e2fc5a0',  # EFI Capsule
    '4a3ca68b-7723-48fb-3d80-578cc1fec44d',  # EFI Capsule v2
    '539182b9-abb5-4391-b69a-e3a943f72fcc',  # UEFI Capsule
    '6dcbd5ed-e82d-4c44-bda1-7194199ad92a',  # Firmware Management Capsule
]

FIRMWARE_GUIDED_GUIDS = {
    "LZMA_COMPRESSED":  "ee4e5898-3914-4259-9d6e-dc7bd79403cf",
    "TIANO_COMPRESSED": "a31280ad-481e-41b6-95e8-127f4c984779",
    "FIRMWARE_VOLUME":  "24400798-3807-4a42-b413-a1ecee205dd8",
    #"VOLUME_SECTION":  "367ae684-335d-4671-a16d-899dbfea6b88",
    "STATIC_GUID":      "fc1bcdb0-7d31-49aa-936a-a4600d9dd083"
}

FIRMWARE_FREEFORM_GUIDS = {
    "CHAR_GUID": "059ef06e-c652-4a45-9fbe-5975e369461c"
}

EFI_FILE_TYPES = {
    # http://wiki.phoenix.com/wiki/index.php/EFI_FV_FILETYPE
    0x00: ("unknown",                    "none",        "0x00"),
    0x01: ("raw",                        "raw",         "RAW"),
    0x02: ("freeform",                   "freeform",    "FREEFORM"),
    0x03: ("security core",              "sec",         "SEC"),
    0x04: ("pei core",                   "pei.core",    "PEI_CORE"),
    0x05: ("dxe core",                   "dxe.core",    "DXE_CORE"),
    0x06: ("pei module",                 "peim",        "PEIM"),
    0x07: ("driver",                     "dxe",         "DRIVER"),
    0x08: ("combined pei module/driver", "peim.dxe",    "COMBO_PEIM_DRIVER"),
    0x09: ("application",                "app",         "APPLICATION"),
    0x0a: ("system management",          "smm",         "SMM"),
    0x0b: ("firmware volume image",      "vol",         "FV_IMAGE"),
    0x0c: ("combined smm/driver",        "smm.dxe",     "COMBO_SMM_DRIVER"),
    0x0d: ("smm core",                   "smm.core",    "SMM_CORE"),
    # 0xc0: ("oem min"),
    # 0xdf: ("oem max"),
    0xf0: ("ffs padding",                "pad",         "0xf0")
}

EFI_SECTION_TYPES = {
    0x01: ("Compression",               "compressed",   None),
    0x02: ("Guid Defined",              "guid",         None),
    0x03: ("Disposable",                "disposable",   None),
    0x10: ("PE32 image",                "pe",           "PE32"),
    0x11: ("PE32+ PIC image",           "pic.pe",       "PIC"),
    0x12: ("Terse executable (TE)",     "te",           "TE"),
    0x13: ("DXE dependency expression", "dxe.depex",    "DXE_DEPEX"),
    # Added from previous code (not in Phoenix spec
    0x14: ("Version section",           "version",      "VERSION"),
    0x15: ("User interface name",       "ui",           "UI"),

    0x16: ("IA-32 16-bit image",        "ia32.16bit",   "COMPAT16"),
    0x17: ("Firmware volume image",     "fv",           "FV_IMAGE"),
    # See FdfParser.py in EDKII's GenFds
    0x18: ("Free-form GUID",            "freeform.guid", "SUBTYPE_GUID"),
    0x19: ("Raw",                       "raw",          "RAW"),
    0x1b: ("PEI dependency expression", "pie.depex",    "PEI_DEPEX"),
    0x1c: ("SMM dependency expression", "smm.depex",    "SMM_DEPEX")
}

EFI_COMPRESSION_TYPES = {
    0x00: "",
    # 0x01: "EFI_STANDARD_COMPRESSION",
    0x01: "PI_STD",
    # 0x02: "EFI_CUSTOMIZED_COMPRESSION"
    0x02: "PI_STD"
}

NVRAM_ATTRIBUTES = {
    "RT":         0x01,
    "DESC_ASCII": 0x02,
    "GUID":       0x04,
    "DATA":       0x08,
    "EXTHDR":     0x10,
    "HER":        0x20,
    "AUTHWD":     0x40,
    "VLD":        0x80,
}


class UEFIVariableHeaderType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("StartId", uint16_t),
        ("State", uint8_t),
        ("Reserved", uint8_t),
        ("Attributes", uint32_t),
        ("NameSize", uint32_t),
        ("DataSize", uint32_t),
        ("VendorGuid", guid_t),
    ]


class NVARVariableHeaderType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("StartId", char * 4),  # NVAR
        ("TotalSize", uint16_t),
        ("Reserved", char * 3),
        ("Attributes", uint8_t),
    ]


class EFIVariableStoreType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Signature", uint32_t),
        ("Size", uint32_t),
        ("Format", uint8_t),
        ("State", uint8_t),
        ("Reserved", uint16_t),
        ("Reserved1", uint32_t),
    ]


class VSSHeaderType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("VendorGuid", guid_t),
        ("StartId", uint16_t),
        ("State", uint8_t),
        ("Reserved", uint8_t),
        ("Attributes", uint32_t),
        ("NameSize", uint32_t),
        ("DataSize", uint32_t),
    ]


class VSSNewHeaderType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("VendorGuid", guid_t),
        ("StartId", uint16_t),
        ("State", uint8_t),
        ("Reserved", uint8_t),
        ("Attributes", uint32_t),
        ("Unknown", char * 28),  # ???
        ("NameSize", uint32_t),
        ("DataSize", uint32_t),
    ]


class TLVHeaderType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Tag0", uint8_t),
        ("Tag1", uint8_t),
        ("Size", uint16_t),
    ]


class EVSARecordType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Signature", char * 4),
        ("Unknown", uint32_t),
        ("Length", uint32_t),
        ("Unknown1", uint32_t),
    ]


class GUIDRecordType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("GuidId", uint16_t),
        ("Guid", guid_t),
    ]


class FirmwareVolumeType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Reserved",   char * 16),  # Zeros
        ("Guid",       guid_t),  #
        ("Size",       uint64_t),
        ("Magic",      char * 4),
        ("Attributes", uint8_t),
        ("HeaderSize", uint32_t),
        ("Checksum",   uint16_t),
        ("Reserved2",  uint8_t),
        ("Revision",   uint8_t)
    ]



# -*- coding: utf-8 -*-

# Instructions:
# 1. Download, and setup your edk2
# 2. make -C <edk2>/BaseTools/Source/C
# 3. The needed binaries are at <edk2>/BaseTools/Source/C/bin/
# 4. GenFds  -f Testing/testing.fdf -o Testing -t GCC46 -b DEBUG -a X64 -p
# OvmfPkg/OvmfPkgX64.dsc


from ..uefi import FirmwareVolume, FirmwareFile, FirmwareFileSystemSection
from ..structs.uefi_structs import EFI_FILE_TYPES, EFI_SECTION_TYPES, EFI_COMPRESSION_TYPES
from ..utils import sguid

FV_BUFFER = '''
[FV.FV_{name}]
BlockSize          = 0x10000
FvAlignment        = 16
ERASE_POLARITY     = 1
MEMORY_MAPPED      = TRUE
STICKY_WRITE       = TRUE
LOCK_CAP           = TRUE
LOCK_STATUS        = TRUE
WRITE_DISABLED_CAP = TRUE
WRITE_ENABLED_CAP  = TRUE
WRITE_STATUS       = TRUE
WRITE_LOCK_CAP     = TRUE
WRITE_LOCK_STATUS  = TRUE
READ_DISABLED_CAP  = TRUE
READ_ENABLED_CAP   = TRUE
READ_STATUS        = TRUE
READ_LOCK_CAP      = TRUE
READ_LOCK_STATUS   = TRUE

{files}
'''

FILE_BUFFER = '''
FILE {type} = {guid} {options} {{
{sections}
}}
'''

SECTION_GUIDED_BUFFER = '''\
{ts}SECTION GUIDED {guid} {auth} {{
{subsections}
{ts}}}
'''

SECTION_COMPRESSED_BUFFER = '''\
{ts}SECTION COMPRESS {type} {{
{subsections}
{ts}}}
'''

SECTION_STRING_BUFFER = '''\
{ts}SECTION {type} = "{value}"
'''

SECTION_LEAF_BUFFER = '''\
{ts}SECTION {type} = {value}
'''


def oguid(guid):
    return sguid(guid)


class GeneratorException(Exception):

    def __init__(self, _object):
        message = "Cannot generate from unsupported type (%s)." % type(_object)
        Exception.__init__(self, message)


class RawSectionGenerator(object):

    def __init__(self, path, ts=''):
        self.embedded = []
        template = SECTION_LEAF_BUFFER

        self.output = template.format(
            ts=ts,
            type="RAW",
            value=path
        )
        pass


class SectionGenerator(object):

    def __init__(self, firmware_section=None, ts=''):
        self.subsections = []
        self.embedded = []
        self.ts = ts

        if firmware_section is not None:
            if not isinstance(firmware_section, FirmwareFileSystemSection):
                raise GeneratorException(firmware_section)
            self._generate(firmware_section)
        pass

    def generate_leaf(self, firmware_section):
        template = SECTION_LEAF_BUFFER
        value = firmware_section.path

        if firmware_section.type == 0x17:  # firmware volume
            name = oguid(firmware_section.guid).replace("-", "_")
            self.embedded.append(
                FirmwareVolumeGenerator(firmware_section.parsed_object, name))
            value = "FV_%s" % name
            pass

        if firmware_section.type in [0x14, 0x15]:
            template = SECTION_STRING_BUFFER
            value = firmware_section.name

        self.output = template.format(
            ts=self.ts,
            type=EFI_SECTION_TYPES[firmware_section.type][2],
            value="%s" % value
        )

    def generate_compressed(self, firmware_section):
        template = SECTION_COMPRESSED_BUFFER

        for subsection in firmware_section.parsed_object.subsections:
            self.subsections.append(
                SectionGenerator(subsection, ts="  %s" % self.ts))

        subsection_output = ""
        for subsection in self.subsections:
            for embedded in subsection.embedded:
                self.embedded.append(embedded)
            subsection_output += "%s" % subsection.output

        self.output = template.format(
            ts=self.ts,
            type=EFI_COMPRESSION_TYPES[firmware_section.parsed_object.type],
            subsections=subsection_output
        )

    def generate_guided(self, firmware_section):
        template = SECTION_GUIDED_BUFFER

        for subsection in firmware_section.parsed_object.subsections:
            self.subsections.append(
                SectionGenerator(subsection, ts="  %s" % self.ts))

        subsection_output = ""
        for subsection in self.subsections:
            for embedded in subsection.embedded:
                self.embedded.append(embedded)
            subsection_output += "%s" % subsection.output

        self.output = template.format(
            ts=self.ts,
            guid=oguid(firmware_section.guid),
            auth="",
            subsections=subsection_output
        )

    def _generate(self, firmware_section):
        if firmware_section.type == 0x01:  # compressed
            self.generate_compressed(firmware_section)
        elif firmware_section.type == 0x02:  # guided
            self.generate_guided(firmware_section)
        else:
            self.generate_leaf(firmware_section)
        pass


class FirmwareFileGenerator(object):

    def __init__(self, firmware_file=None, type_label=None, guid=None):
        self.embedded = []
        self.sections = []

        self.type_label = type
        self.guid_label = guid
        if firmware_file is not None:
            if not isinstance(firmware_file, FirmwareFile):
                raise GeneratorException(firmware_file)
            self._generate(firmware_file)
        pass

    def add_section(self, firmware_section):
        self.sections.append(SectionGenerator(firmware_section, ts='  '))

    def _generate(self, firmware_file):
        self.output = ""

        if firmware_file.type == 0xf0:
            # Do not add FFS padding files
            return

        self.type_label = EFI_FILE_TYPES[firmware_file.type][2]
        self.guid_label = oguid(firmware_file.guid)

        if firmware_file.type == 0x01:
            #self.sections = [RawSectionGenerator(blob) for blob in firmware_file.raw_blobs]
            self.sections.append(
                RawSectionGenerator(firmware_file.path, ts='  '))
        else:
            for section in firmware_file.sections:
                self.add_section(section)

        section_output = ""
        for section in self.sections:
            section_output += "%s" % section.output
            for embedded in section.embedded:
                #self.output += embedded.output
                self.embedded.append(embedded)

        template = FILE_BUFFER
        self.output = template.format(
            type=self.type_label,
            guid=self.guid_label,
            options="",
            sections=section_output
        )
        pass


class FirmwareVolumeGenerator(object):

    def __init__(self, volume=None, name="GENERIC"):
        self.files = []

        self.name = name
        if volume is not None:
            if not isinstance(volume, FirmwareVolume):
                raise GeneratorException(volume)
            self._generate(volume)
        pass

    def add_file(self, firmware_file):
        self.files.append(FirmwareFileGenerator(firmware_file))

    def _generate(self, volume):
        '''Generate buffer using volume object.'''
        self.output = ""

        # Do not generate separate block sets
        for filesystem in volume.firmware_filesystems:
            for firmware_file in filesystem.files:
                self.add_file(firmware_file)

        file_output = ""
        for firmware_file in self.files:
            for embedded in firmware_file.embedded:
                self.output += embedded.output
            file_output += "%s" % firmware_file.output
        template = FV_BUFFER
        self.output += template.format(
            name=self.name,
            files=file_output
        )



'''Base provides basic firmware object structures.
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import ctypes

from .utils import dump_data, sguid, blue


class BaseObject(object):
    '''A base object can be used to access direct content.'''


class FirmwareObject(object):
    '''A pseudo-abstract type providing common firmware member facilities.'''
    def __init__(self):
        self.data = None
        self.name = None
        self.attrs = None
        self.guid = None

    @property
    def content(self):
        '''The object content is the 'data' stream.'''
        if hasattr(self, "data") and self.data is not None:
            return self.data
        return ""

    @property
    def objects(self):
        '''Objects are the child firmware objects found via 'processing'.'''
        return []

    @property
    def label(self):
        '''An overload for an object 'name'.'''
        if hasattr(self, "name") and self.name is not None:
            if self.name is None:
                return ""
            return self.name
        return ""

    @property
    def guid_label(self):
        '''A string representation of an optional 'guid' field.'''
        if not hasattr(self, "guid") or self.guid is None:
            return ""
        return sguid(self.guid)

    @property
    def type_label(self):
        '''The string representation of the object's class name.'''
        return self.__class__.__name__

    @property
    def attrs_label(self):
        '''An overload for the 'attrs' field.'''
        if hasattr(self, "attrs") and self.attrs is not None:
            return self.attrs
        return {}

    def info(self, include_content=False):
        '''Firmwae objects define a common interface for information.

        This defines: label, guid, type, content, attrs-- as common between
        most firmware objects.

        Args:
            include_content (Optional[bool]): Include a pointer to the 'data'
            or content stream.

        Return:
            dict: Return a pointer to this object "_self" and the defines listed
                above with an optional pointer to the data stream.
        '''
        return {
            "_self": self,
            "label": self.label,
            "guid": self.guid_label,
            "type": self.type_label,
            "content": self.content if include_content else "",
            "attrs": self.attrs_label
        }

    def iterate_objects(self, include_content=False):
        '''Flatten this object's children into a list.

        This mis-named as an interation. Each object within the children list
        is recursively 'iterated', meaning its 'iterate_objects' method is
        called. The object is represented via the 'info' method. Access to the
        object is possible via the "_self" key.

        The output list does not include this object but each entry sets a
        "_parent" key with a pointer to this object.

        Return:
            list: flattened list of firmware objects.
        '''
        objects = []
        for _object in self.objects:
            if _object is None:
                continue
            _info = _object.info(include_content)
            _info["objects"] = _object.iterate_objects(include_content)
            for _object in _info["objects"]:
                _object["parent"] = _info
            objects.append(_info)
        return objects


class StructuredObject(object):
    def __init__(self):
        self.fields = []

    def parse_structure(self, data, structure):
        '''Construct an instance object of the provided structure.'''
        struct_instance = structure()
        struct_size = ctypes.sizeof(struct_instance)

        struct_data = data[:struct_size]
        struct_length = min(len(struct_data), struct_size)
        ctypes.memmove(
            ctypes.addressof(struct_instance), struct_data, struct_length)
        self.structure = struct_instance
        self.structure_data = struct_data
        self.structure_fields = [field[0] for field in structure._fields_]
        self.structure_size = struct_size

    def show_structure(self):
        for field in self.fields:
            print ("%s: %s" % (field, getattr(self.structure, field, None)))


class RawObject(FirmwareObject, BaseObject):

    def __init__(self, data):
        self.data = data

    def build(self, generate_checksum, debug=False):
        return self.data

    def showinfo(self, ts='', index=None):
        print ("%s%s size= %d " % (
            ts, blue("RawObject:"), len(self.data)
        ))

    def dump(self, parent='', index=None):
        path = os.path.join(parent, "object.raw")
        dump_data(path, self.data)


class AutoRawObject(RawObject):
    '''A RawObject that applies AutoParser logic for embedded object discovery.
    '''

    def __init__(self, data):
        self.object = None
        self.data = data

    @property
    def objects(self):
        if self.object is not None:
            return [self.object]
        return []

    def process(self):
        from . import AutoParser
        parser = AutoParser(self.data)
        self.object = parser.parse()

    def showinfo(self, ts='', index=None):
        if self.object is None:
            print ("%s%s size= %d " % (
                ts, blue("RawObject:"), len(self.data)
            ))
            return
        self.object.showinfo(ts)

    def dump(self, parent='', index=None):
        if self.object is None:
            path = os.path.join(parent, "object.raw")
            dump_data(path, self.data)
            return
        self.object.dump(parent)



'''UEFI Firmware parser utils.
'''

import uefi
import pfs
import me
import flash

from misc import checker
from base import FirmwareObject
from utils import search_firmware_volumes


class AutoParser(object):
    '''Automatically detect the file format type.

    The UEFI Firmware module supports several types of file formats:
      - UEFI Firmware Volumes
      - (U)EFI Capsule Updates
      - UEFI Firmware Files
      - Dell "PFS" Files
      - Intel ME Modules
      - Flash Descriptors and Regions

    If the input file format is unknown the AutoParser can attempt to discover
    the type by applying basic checks for known headers.
    '''

    def __init__(self, data, search=True):
        '''Create an AutoParser instance.

        Args:
            data (binary): The entire input file contents.
            search (Optional[bool]): Allow brute-force discovery of volumes.
        '''
        self.data = data
        self.data_type = 'unknown'
        self.constructor = None
        self.firmware = None

        header = data[:100]
        for tester in checker.TESTERS:
            if tester().match(header):
                self.data_type = tester().name
                self.constructor = tester().parser

    def type(self):
        '''Return the discovered file format type.

        If the type could not be matched, 'unknown' is returned.

        Return:
            string: Name of the file format type.
        '''
        return self.data_type

    def parse(self):
        '''Call the 'process' method for the discovered type using the input
        file contents. If the file type's parser returns False indicating a
        failure or exception while parsing this will return None.

        Return:
            object: The associated file object upon success, otherwise None.
        '''
        if self.constructor is None:
            return None
        if self.firmware is not None:
            return self.firmware

        # Instanciate an instance of the firmware object
        self.firmware = self.constructor(self.data)
        if not self.firmware.process():
            # Parsing failed, remove the object reference.
            self.firmware = None
            return None
        if self.constructor is uefi.FirmwareVolume:
            mfc = MultiVolumeContainer(self.data[self.firmware.size:])
            if mfc.has_indexes():
                # Headers were discovered, attempt to process.
                if mfc.process():
                    # Add the base (first) firmware volume
                    mfc.append_base(self.firmware)
                self.firmware = mfc
        return self.firmware


class MultiVolumeContainer(FirmwareObject):
    '''The AutoParser attempts to search for 'stacked' UEFIFirmwareVolumes.

    By default if a volume does not consist of the entire input content the
    AutoParser will attempt to search for additional volume headers. This can
    be disabled with the associated 'search' optional parameter.

    If additional volumes are discovered this MultiVolumeContainer is returned
    instead of a UEFIFirmwareVolume.
    '''

    def __init__(self, data):
        '''Initialize the container with the tail content from a volume.'''
        self.data = data
        self.indexes = search_firmware_volumes(data)
        self.volumes = []

    def has_indexes(self):
        '''Check if any indexes were discovered.'''
        return self.indexes > 0

    def append_base(self, volume):
        '''Set the base volume as the first within the list.'''
        self.volumes = [volume] + self.volumes

    def process(self):
        for index in self.indexes:
            volume = uefi.FirmwareVolume(self.data[index - 40:], index)
            if volume.process():
                self.volumes.append(volume)
        return len(self.volumes) > 0

    @property
    def objects(self):
        return self.volumes

    def showinfo(self, ts='', index=None):
        '''Write structure information to stdout.'''
        for volume in self.volumes:
            volume.showinfo(ts, index)

    def dump(self, parent, index=None):
        '''Allow a caller to dump the content of volumes.'''
        for i in range(self.volumes):
            path = os.path.join(parent, "volume-%d" % i)
            self.volumes[i].dump(path)


__title__ = "uefi_firmware"
__version__ = "1.6"
__author__ = "Teddy Reed"
__license__ = "BSD"



import efiguids
import efiguids_ami
import efiguids_dell
import efiguids_lenovo
import efiguids_asrock

from ..utils import aguid

GUID_TABLES = [
    efiguids.GUIDs,
    efiguids_ami.GUIDs,
    efiguids_dell.GUIDs,
    efiguids_lenovo.GUIDs,
    efiguids_asrock.GUIDs,
]


def get_guid_name(guid):
    raw_guid = aguid(guid) if isinstance(guid, str) else guid

    for guid_table in GUID_TABLES:
        for name, match_guid in guid_table.iteritems():
            match = True
            for i, k in enumerate(raw_guid):
                if match_guid[i] != k:
                    match = False
                    break
            if match:
                return name
    return None


def get_tables():
    return GUID_TABLES



import ctypes
import struct

uint8_t = ctypes.c_ubyte
char = ctypes.c_char
uint32_t = ctypes.c_uint
uint64_t = ctypes.c_uint64
uint16_t = ctypes.c_ushort

ME_HEADER = "\x20\x20\x80\x0F\x40\x00\x00\x24"
ME_PARTITION_HEADER = "$FPT"

def replace_bad(value, deletechars):
    for c in deletechars:
        value = value.replace(c, '_')
    return value


def read_struct(li, struct):
    s = struct()
    slen = ctypes.sizeof(s)
    _bytes = li.read(slen)
    fit = min(len(_bytes), slen)
    ctypes.memmove(ctypes.addressof(s), _bytes, fit)
    return s


def get_struct(str_, off, struct):
    s = struct()
    slen = ctypes.sizeof(s)
    _bytes = str_[off:off + slen]
    fit = min(len(_bytes), slen)
    ctypes.memmove(ctypes.addressof(s), _bytes, fit)
    return s


def DwordAt(f, off):
    return struct.unpack("<I", f[off:off + 4])[0]


class MeModuleFileHeader1Type(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Tag",            char * 4),  # $MOD
        ("Unk04",          uint32_t),
        ("Unk08",          uint32_t),
        ("MajorVersion",   uint16_t),
        ("MinorVersion",   uint16_t),
        ("HotfixVersion",  uint16_t),
        ("BuildVersion",   uint16_t),
        ("Unk14",          uint32_t),
        ("CompressedSize", uint32_t),
        ("UncompressedSize", uint32_t),
        ("LoadAddress",    uint32_t),
        ("MappedSize",     uint32_t),
        ("Unk28",          uint32_t),
        ("Unk2C",          uint32_t),
        ("Name",           char * 16),
        ("Guid",           uint8_t * 16),
    ]


class MeModuleHeader1Type(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Tag",            char * 4),  # $MME
        ("Guid",           uint8_t * 16),
        ("MajorVersion",   uint16_t),
        ("MinorVersion",   uint16_t),
        ("HotfixVersion",  uint16_t),
        ("BuildVersion",   uint16_t),
        ("Name",           char * 16),
        ("Hash",           uint8_t * 20),
        ("Size",           uint32_t),
        ("Flags",          uint32_t),
        ("Unk48",          uint32_t),
        ("Unk4C",          uint32_t),
    ]
    udc_format = "<4s20s16sII"
    udc_length = 0x30


class MeModuleHeader2Type(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Tag",            char * 4),  # $MME
        ("Name",           char * 16),
        ("Hash",           uint8_t * 32),
        ("Unk34",          uint32_t),
        ("Offset",         uint32_t),  # From the manifest
        ("Unk3C",          uint32_t),
        ("Size",           uint32_t),
        ("Unk44",          uint32_t),
        ("Unk48",          uint32_t),
        ("LoadBase",       uint32_t),
        ("Flags",          uint32_t),
        ("Unk54",          uint32_t),
        ("Unk58",          uint32_t),
        ("Unk5C",          uint32_t),
    ]
    udc_format = "<4s32s16sII"
    udc_length = 0x3c


class HuffmanLUTHeader(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Tag",            char * 4),  # LLUT
        ("ChunkCount",     uint32_t),
        ("DecompBase",     uint32_t),
        ("Unk0C",          uint32_t),
        ("Size",           uint32_t),
        ("DataStart",      uint32_t),  # Start of data
        ("Unk18",          uint32_t * 6),
        ("ChunkSize",      uint32_t),
        ("Unk34",          uint32_t),
        ("Chipset",        char * 8),  # PCH
    ]


class HuffmanOffsetBytes(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Offset", uint32_t, 24),
        ("Length", uint8_t),
    ]


class HuffmanOffsets(ctypes.Union):
    _fields_ = [
        ("b", HuffmanOffsetBytes),
        ("asword", uint32_t),
    ]


class MeManifestHeaderType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("ModuleType",     uint16_t),  # 00
        ("ModuleSubType",  uint16_t),  # 02
        ("HeaderLen",      uint32_t),  # 04 in dwords
        ("HeaderVersion",  uint32_t),  # 08
        ("Flags",          uint32_t),  # 0C 0x80000000 = Debug
        ("ModuleVendor",   uint32_t),  # 10
        ("Date",           uint32_t),  # 14 BCD yyyy.mm.dd
        ("Size",           uint32_t),  # 18 in dwords
        ("Tag",            char * 4),  # 1C $MAN or $MN2
        ("NumModules",     uint32_t),  # 20
        ("MajorVersion",   uint16_t),  # 24
        ("MinorVersion",   uint16_t),  # 26
        ("HotfixVersion",  uint16_t),  # 28
        ("BuildVersion",   uint16_t),  # 2A
        ("Unknown1",       uint32_t * 19),  # 2C
        ("KeySize",        uint32_t),  # 78
        ("ScratchSize",    uint32_t),  # 7C
        ("RsaPubKey",      uint32_t * 64),  # 80
        ("RsaPubExp",      uint32_t),  # 180
        ("RsaSig",         uint32_t * 64),  # 184
        ("PartitionName",  char * 12),  # 284
        # 290
    ]


class MePartitionTable(ctypes.LittleEndianStructure):
    _fields_ = [
        ("_blank",          char * 16),
        ("Magic",           char * 4), # $FPT
        ("Entries",         uint32_t),
        ("Version",         uint8_t),
        ("Type",            uint8_t),
        ("Size",            uint8_t),
        ("Checksum",        uint8_t),
        ("LIFE",            uint16_t),
        ("LIM",             uint16_t),
        ("UMASize",         uint32_t),
        ("Flags",           uint32_t),
        ("Unknown1",         char * 8),
    ]


class MeFptEntryType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("Name",            char * 4),  # 00 partition name
        ("Owner",           char * 4),  # 04 partition owner?
        ("Offset",          uint32_t),  # 08 from the start of FPT, or 0
        ("Size",            uint32_t),  # 0C
        ("TokensOnStart",   uint32_t),  # 10
        ("MaxTokens",       uint32_t),  # 14
        ("ScratchSectors",  uint32_t),  # 18
        ("Flags",           uint32_t),  # 1C
    ]


class AcManifestHeaderType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("ModuleType",     uint16_t),  # 00
        ("ModuleSubType",  uint16_t),  # 02
        ("HeaderLen",      uint32_t),  # 04 in dwords
        ("HeaderVersion",  uint32_t),  # 08
        ("ChipsetID",      uint16_t),  # 0C
        ("Flags",          uint16_t),  # 0E 0x80000000 = Debug
        ("ModuleVendor",   uint32_t),  # 10
        ("Date",           uint32_t),  # 14 BCD yyyy.mm.dd
        ("Size",           uint32_t),  # 18 in dwords
        ("Reserved1",      uint32_t),  # 1C
        ("CodeControl",    uint32_t),  # 20
        ("ErrorEntryPoint", uint32_t),  # 24
        ("GDTLimit",       uint32_t),  # 28
        ("GDTBasePtr",     uint32_t),  # 2C
        ("SegSel",         uint32_t),  # 30
        ("EntryPoint",     uint32_t),  # 34
        ("Reserved2",      uint32_t * 16),  # 38
        ("KeySize",        uint32_t),  # 78
        ("ScratchSize",    uint32_t),  # 7C
        ("RsaPubKey",      uint32_t * 64),  # 80
        ("RsaPubExp",      uint32_t),  # 180
        ("RsaSig",         uint32_t * 64),  # 184
        # 284
    ]


class MeCpdEntryType(ctypes.LittleEndianStructure):
    size = 0x18

    _fields_ = [
        ("Name",            char * 12),
        ("Offset",          uint32_t),
        ("Size",            uint32_t),
        ("Flags",           uint32_t),
    ]


class MeCpdHeaderType(ctypes.LittleEndianStructure):
    size = 0x10

    _fields_ = [
        ("Tag",             char * 4),
        ("NumModules",      uint32_t),
        ("Flags",           uint32_t),
        ("PartitionName",   char * 4),
    ]



# -*- coding: utf-8 -*-

import os
import sys
import struct


def blue(msg):
    '''Return the input string as console-escaped blue.'''
    return "\033[1;36m%s\033[1;m" % msg


def red(msg):
    '''Return the input string as console-escaped red.'''
    return "\033[31m%s\033[1;m" % msg


def green(msg):
    '''Return the input string as console-escaped green.'''
    return "\033[32m%s\033[1;m" % msg


def purple(msg):
    '''Return the input string as console-escaped purple.'''
    return "\033[1;35m%s\033[1;m" % msg


def print_error(msg):
    '''Write the input string to stderr.'''
    print >> sys.stderr, msg


def ascii_char(c):
    '''Return the ASCII or (.) representation of the input character.'''
    if ord(c) >= 32 and ord(c) <= 126:
        return c
    return '.'


def hex_dump(data, size=16):
    '''Print a debug view of binary data similar to a hex editor

    Args:
        data (binary): Data to be printed.
        size (Optional[int]): Length of each line.
    '''
    def print_line(line):
        print "%s | %s" % (
            line.encode("hex"),
            "".join([ascii_char(c) for c in line])
        )
        pass

    for i in xrange(0, len(data) / size):
        data_line = data[i * size:i * size + size]
        print_line(data_line)

    if not len(data) % size == 0:
        print_line(data[(len(data) % size) * -1:])


def sguid(b, big=False):
    '''RFC4122 binary GUID as string.'''
    if b is None or len(b) != 16:
        return ""
    a, b, c, d = struct.unpack("%sIHH8s" % (">" if big else "<"), b)
    d = ''.join('%02x' % ord(c) for c in d)
    return "%08x-%04x-%04x-%s-%s" % (a, b, c, d[:4], d[4:])


def s2aguid(s):
    '''RFC4122 string GUID as int array.'''
    guid = [s[:8], s[8 + 1:9 + 4], s[13 + 1:14 + 4],
            s[18 + 1:19 + 4] + s[-12:]]
    return aguid("".join([part.decode("hex") for part in guid]), True)


def a2sguid(a):
    '''RFC4122 int array GUID as string.'''
    guid = ""
    for value in a:
        value = format(value, 'x')
        guid += value.zfill(len(value) + len(value) % 2).decode("hex")
    guid = guid.zfill(len(guid) + len(guid) % 2)
    return sguid(guid, True)


def aguid(b, big=False):
    '''RFC4122 binary GUID as int array.'''
    a, b, c, d = struct.unpack("%sIHH8s" % (">" if big else "<"), b)
    return [a, b, c] + [ord(_c) for _c in d]


def bit_set(field, bit):
    '''Check if bit is set (1) in field.'''
    return (field & bit == bit)


def dump_data(name, data):
    '''Write binary data to name.

    Args:
        name (string): Path to output file, created if it does not exist.
        data (binary): Content to be written.
    '''
    try:
        if os.path.dirname(name) is not '':
            if not os.path.exists(os.path.dirname(name)):
                os.makedirs(os.path.dirname(name))
        with open(name, 'wb') as fh:
            fh.write(data)
        print "Wrote: %s" % (red(name))
    except Exception as e:
        print "Error: could not write (%s), (%s)." % (name, str(e))


def search_firmware_volumes(data, byte_align=16, limit=None):
    '''"Search a blob for '_FVH' magics, related to firmware volume headers.'''
    potential_volumes = []
    for aligned in xrange(32, len(data), byte_align):
        if data[aligned:aligned + 4] == '_FVH':
            potential_volumes.append(aligned)
            if limit and limit == len(potential_volumes):
                return potential_volumes
        magic = data[(aligned + byte_align / 2):(aligned + byte_align / 2 + 4)]
        if magic == '_FVH':
            potential_volumes.append(aligned + byte_align / 2)
            if limit and limit == len(potential_volumes):
                return potential_volumes
    return potential_volumes
    pass


def flatten_firmware_objects(base_objects):
    '''Flatten the parent-child relations between firmware objects.

    It is often nice to remove the structured relationship between firmware
    objects. Each object maintains a pointer to its parent and will
    maintain pointers to children. The output will contain a flattened list
    of all children.

    Args:
        base_objects (list): Nested list of firmware objects.

    Returns:
        list: Non-Nested list of firmware objects.
    '''
    objects = []
    for _object in base_objects:
        objects.append(_object)
        if "objects" in _object and len(_object["objects"]) > 0:
            objects += flatten_firmware_objects(_object["objects"])
    return objects



import ctypes

FLASH_HEADER = "\x5A\xA5\xF0\x0F"

uint8_t = ctypes.c_ubyte
char = ctypes.c_char
uint32_t = ctypes.c_uint
uint64_t = ctypes.c_uint64
uint16_t = ctypes.c_ushort


class FlashDescriptorMapType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("ComponentBase",       uint8_t),  #
        ("NumberOfFlashChips",  uint8_t),  #
        ("RegionBase",          uint8_t),  #
        ("NumberOfRegions",     uint8_t),  #
        ("MasterBase",          uint8_t),  #
        ("NumberOfMasters",     uint8_t),  #
        ("PchStrapsBase",       uint8_t),  #
        ("NumberOfPchStraps",   uint8_t),  #
        ("ProcStrapsBase",      uint8_t),  #
        ("NumberOfProcStraps",  uint8_t),  #
        ("IccTableBase",        uint8_t),  #
        ("NumberOfIccTableEntries", uint8_t),  #
        ("DmiTableBase",            uint8_t),  #
        ("NumberOfDmiTableEntries", uint8_t),  #
        ("ReservedZero",            uint16_t),  #
    ]


class FlashMasterSectionType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("BiosId",    uint16_t),  #
        ("BiosRead",  uint8_t),   #
        ("BiosWrite", uint8_t),   #
        ("MeId",      uint16_t),  #
        ("MeRead",    uint8_t),   #
        ("MeWrite",   uint8_t),   #
        ("GbeId",     uint16_t),  #
        ("GbeRead",   uint8_t),   #
        ("GbeWrite",  uint8_t),   #
    ]


class FlashRegionSectionType(ctypes.LittleEndianStructure):
    _fields_ = [
        ("ReservedZero",        uint16_t),  #
        ("FlashBlockEraseSize", uint16_t),  #
        ("BiosBase",            uint16_t),  #
        ("BiosLimit",           uint16_t),  #
        ("MeBase",              uint16_t),  #
        ("MeLimit",             uint16_t),  #
        ("GbeBase",             uint16_t),  #
        ("GbeLimit",            uint16_t),  #
        ("PdrBase",             uint16_t),  #
        ("PdrLimit",            uint16_t),  #
    ]



'''
  Modified version 2014-01-10 Teddy Reed
  Modified version 2013-12-29 Damien Zammit

  Based on the original work from:
    Intel ME ROM image dumper/extractor
    Copyright (c) 2012 Igor Skochinsky
    Version 0.1 2012-10-10
    Version 0.2 2013-08-15

    This software is provided 'as-is', without any express or implied warranty.
    In no event will the authors be held liable for any damages arising from the
    use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software in a
    product, an acknowledgment in the product documentation would be appreciated
    but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
'''

import ctypes
import struct
import os
import array

from .structs.intel_me_structs import *
from .utils import *
from uefi_firmware import efi_compressor
from uefi_firmware.base import FirmwareObject, StructuredObject

MeModulePowerTypes = ["POWER_TYPE_RESERVED",
                      "POWER_TYPE_M0_ONLY", "POWER_TYPE_M3_ONLY", "POWER_TYPE_LIVE"]
MeCompressionTypes = ["COMP_TYPE_NOT_COMPRESSED",
                      "COMP_TYPE_HUFFMAN", "COMP_TYPE_LZMA", "<unknown>"]
COMP_TYPE_NOT_COMPRESSED = 0
COMP_TYPE_HUFFMAN = 1
COMP_TYPE_LZMA = 2
MeModuleTypes = ["DEFAULT", "PRE_ME_KERNEL",
                 "VENOM_TPM", "APPS_QST_DT", "APPS_AMT", "TEST"]
MeApiTypes = ["API_TYPE_DATA", "API_TYPE_ROMAPI",
              "API_TYPE_KERNEL", "<unknown>"]


class MeObject(StructuredObject, FirmwareObject):

    '''An ME Object is a combination of a parsing/extraction class and a ctype
    definding structure object.

    This follows the same ctor, process, showinfo, dump calling convention.
    '''

    def show_compression(self):
        if self.compression == COMP_TYPE_HUFFMAN:
            print " (huffman)"
        elif self.compression == COMP_TYPE_LZMA:
            print " (lzma)"
        else:
            print ""

    def dump_module(self, parent):
        if self.compression == COMP_TYPE_LZMA:
            dump_data("%s.module.lzma" %
                      os.path.join(parent, self.name), self.data)
            try:
                data = efi_compressor.LzmaDecompress(self.data, len(self.data))
                dump_data("%s.module" % os.path.join(parent, self.name), data)
            except Exception as e:
                print "Cannot extract (%s), %s" % (self.name, str(e))
                return
        elif self.compression == COMP_TYPE_NOT_COMPRESSED:
            dump_data("%s.module" % os.path.join(parent, self.name), self.data)


class MeModule(MeObject):

    def __init__(self, data, structure_type, offset):
        self.attrs = {}
        self.parse_structure(data, structure_type)
        self.structure_type = structure_type

        if structure_type == MeModuleHeader2Type:
            self.structure.Guid = "(none)"
            self.attrs["version"] = "0.0.0.0"
        elif structure_type == MeModuleHeader1Type:
            self.attrs["version"] = "%d.%d.%d.%d" % (
                self.structure.MajorVersion, self.structure.MinorVersion,
                self.structure.HotfixVersion,
                self.structure.BuildVersion
            )

        self.guid = self.structure.Guid
        self.name = self.structure.Name
        self.tag = self.structure.Tag

        self.attrs["module_size"] = self.structure.Size
        if structure_type == MeModuleHeader1Type:
            self.attrs["load_base"] = 0
            self.offset = offset
        elif structure_type == MeModuleHeader2Type:
            self.attrs["load_base"] = self.structure.LoadBase
            # Must know the offset from given data (the start of the header) to find data
            self.offset = self.structure.Offset - offset

        self.attrs["flags"] = self.structure.Flags
        if structure_type == MeModuleHeader2Type:
            self.attrs["power_type"] = (
                self.structure.Flags >> 1) & 3  # MeModulePowerTypes
            self.attrs["compression"] = (
                self.structure.Flags >> 4) & 7  # MeCompressionTypes
            # MeModuleTypes (optional)
            self.attrs["module_stage"] = (self.structure.Flags >> 7) & 0xF
            self.attrs["api_type"] = (
                self.structure.Flags >> 11) & 7  # MeApiTypes
            self.attrs["privileged"] = ((self.structure.Flags >> 16) & 1)
        # There are unknown flags to parse, todo: revisit
        pass

        # print "Debug: module data 0x%08X - 0x%08X" % (
        #    self.offset, self.offset + self.structure.Size)
        self.data = data[self.offset:self.offset + self.structure.Size]

    def process(self):
        if self.compression == COMP_TYPE_HUFFMAN:
            # The individual modules are compressed together in a partition
            # chunk
            return True

        if self.structure_type == MeModuleHeader1Type:
            # It's possible for type 1 to include LZMA compression
            if self.data[0x50:0x55] == '\x5D\x00\x00\x80\x00':
                raw_data = self.data[0x50:0x55]
                raw_data += struct.pack("<Q", self.structure.UncompressedSize)
                raw_data += self.data[0x55:]
                self.data = raw_data
        return True

    @property
    def size(self):
        return ctypes.sizeof(self.structure_type)

    @property
    def compression(self):
        if self.structure_type == MeModuleHeader1Type:
            return COMP_TYPE_NOT_COMPRESSED
        else:
            return (self.structure.Flags >> 4) & 7

    def showinfo(self, ts=''):
        guid = self.guid
        if self.guid != "(none)":
            guid = green(sguid(self.guid))
        print "%s%s name= %s, guid= %s, version= %s, size= %s" % (
            ts, blue("ME Module"),
            purple(self.name),
            guid, self.attrs["version"], self.attrs["module_size"]),
        self.show_compression()

    def dump(self, parent=""):
        self.dump_module(parent)


class MeVariableModule(MeObject):
    HEADER_SIZE = 8

    def __init__(self, data, structure_type):
        self.update = {}
        self.type = structure_type
        self.data = None
        self.size = 0

        hdr = data[:self.HEADER_SIZE]
        self.valid_header = True
        self.header_blank = False

        if hdr == '\xFF' * self.HEADER_SIZE:
            self.header_blank = True
            return

        if len(hdr) < self.HEADER_SIZE or hdr[0] != '$':
            # print "Debug: invalid module header."
            self.valid_header = False
            return

        self.tag = hdr[:4]
        # Note the elen size includes the header size
        self.size = struct.unpack("<I", hdr[4:])[0] * 4 - self.HEADER_SIZE
        self.data = data[self.HEADER_SIZE:self.HEADER_SIZE + self.size]
        pass

    def add_update(self, tag, name, offset, size):
        self.update["tag"] = tag
        self.update["name"] = name
        self.update["offset"] = offset
        self.update["size"] = size
        pass

    def process(self):
        if self.tag == '$UDC':
            subtag, _hash, name, offset, size = struct.unpack(
                self.type.udc_format, self.data[:self.type.udc_length])
            self.add_update(subtag, name, offset, size)
        if self.tag in ['$SKU', '$UVR']:
            # SKU is not handled
            self.values = [0, 0]
            return True
        if self.size == 3:
            values = [struct.unpack("<I", self.data[:4])[0]]
        if self.size == 4:
            values = struct.unpack("<II", self.data[:8])[0]
        else:
            values = array.array("I", self.data)

        self.values = values
        return True

    def showinfo(self, ts=''):
        print "%s%s tag= %s, size= %d" % (
            ts, blue("VModule"), purple(self.tag), self.size)
        if self.tag == '$UDC':
            print "%s%s name= %s, offset= %d, size= %s" % (
                ts, blue("%s Update" % self.update["tag"]),
                purple(self.update["name"]),
                self.update["offset"], self.update["size"])
        pass


class MeModuleFile(MeObject):

    def __init__(self, data):
        self.size = 0
        tag = data[:4]

        self.valid_header = True
        if tag != "$MOD":
            self.valid_header = False
            return

        self.parse_structure(data, MeModuleFileHeader1Type)
        self.name = self.structure.Name.rstrip('\0')
        self.size = self.structure.Size
        pass


class MeLLUT(MeObject):

    def __init__(self, data, relative_offset):
        #self.tag = data[:4]

        self.valid_header = True
        if data[:4] != 'LLUT':
            self.valid_header = False

        #hdr = data[4:52]
        # chunkcount, decompbase, unk0c, size, start, a,b,c,d,e,f, chunksize = struct.unpack(
        #    "<IIIIIIIIIIII", hdr)
        self.parse_structure(data, HuffmanLUTHeader)
        self.size = self.structure.Size

        # The start and end addresses are relative to the manifest.
        # The relative offset references the start of the manifest data (not
        # header).
        self.start = self.structure.DataStart
        self.offset = relative_offset

        self.chunkcount = self.structure.ChunkCount
        self.chunksize = self.structure.ChunkSize
        self.decompression_base = self.structure.DecompBase

        self.data = data[
            self.start - relative_offset:self.start - relative_offset + self.size]
        # The huffman look up table is stored following the header data.
        self.lut_data = data[
            self.structure_size:self.structure_size + self.chunkcount * 4]

    def showinfo(self, ts=''):
        print "%s%s chunks= %d, chunk size= %d, start= %d, size= %d, base= 0x%08X" % (
            ts, blue("LLUT"),
            self.chunkcount, self.chunksize, self.start, self.size, self.decompression_base)

    def dump(self, parent='PART'):
        # print "Debug: relative (%d) absolute start (%d) len (%d)." % (
        #    self.offset, self.start, len(self.data))
        dump_data("%s.llut.table" % parent, self.lut_data)
        dump_data("%s.llut.compressed" % parent, self.data)


class MeManifestHeader(MeObject):
    _DATA_OFFSET = 12

    def __init__(self, data, container_offset=0):
        self.attrs = {}

        self.valid_header = True
        if data[:8] != "\x04\x00\x00\x00\xA1\x00\x00\x00":
            self.valid_header = False
            return

        self.parse_structure(data, MeManifestHeaderType)
        # Save the container offset as LLUT start is an absolute reference
        self.container_offset = container_offset
        self.data = data[self.partition_offset:]

        '''Set storage attributes.'''
        self.attrs["header_version"] = "%d.%d" % (
            self.structure.HeaderVersion >> 16, self.structure.HeaderVersion & 0xFFFF)
        self.attrs["version"] = "%d.%d.%d.%d" % (
            self.structure.MajorVersion,
            self.structure.MinorVersion,
            self.structure.HotfixVersion,
            self.structure.BuildVersion
        )
        self.attrs["flags"] = "0x%08X" % (self.structure.Flags)
        self.attrs["module_vendor"] = "0x%04X" % (self.structure.ModuleVendor)
        self.attrs["date"] = "%08X" % (self.structure.Date)
        self.size = self.structure.Size

        '''Skipped.'''
        #ModuleType, ModuleSubType, size, tag, num_modules, keysize, scratchsize, rsa

        self.partition_name = self.structure.PartitionName.rstrip("\0")
        if not self.partition_name:
            self.partition_name = "(none)"

        self.modules = []
        self.variable_modules = []
        self.partition_end = 0
        self.huffman_llut = None

    @property
    def absolute_offset(self):
        return self.structure.HeaderLen * 4 + self._DATA_OFFSET + self.container_offset

    @property
    def partition_offset(self):
        return self.structure.HeaderLen * 4 + self._DATA_OFFSET

    @property
    def objects(self):
        _objects = self.modules + self.variable_modules
        if self.huffman_llut is not None:
            _objects.append(self.huffman_llut)
        return _objects

    def showinfo(self, ts=''):
        print "%s%s type= %d, subtype= %d, partition name= %s" % (
            ts, blue("ME Module Manifest"),
            self.structure.ModuleType, self.structure.ModuleSubType,
            purple(self.structure.PartitionName))
        for module in self.modules:
            module.showinfo(ts="  %s" % ts)
        for module in self.variable_modules:
            module.showinfo(ts="  %s" % ts)
        if self.huffman_llut is not None:
            self.huffman_llut.showinfo(ts="  %s" % ts)

    def _parse_mods(self):
        # Parse the module headers (two types of headers, specified by the
        # manifest).
        module_offset = 0
        huffman_offset = 0
        for module_index in xrange(self.structure.NumModules):
            module = MeModule(
                self.data[module_offset:],
                self.header_type, module_offset + self.partition_offset)
            # print "Debug: found me module header (%s) at (%d)." %
            # (module.tag, module_offset)
            if module.compression == COMP_TYPE_HUFFMAN:
                # Todo: skipped precondition for huffman offsets.
                # print "Debug: Setting huffman offset: %d" %
                # module.structure.Offset
                huffman_offset = module.structure.Offset
            if not module.process():
                return False
            self.modules.append(module)
            self.module_map[module.name] = module
            module_offset += module.size

        #additional_header = self.structure.Size*4 - module_offset
        # print "Debug: Remaining header: %d - %d = %d" % (
        #    self.structure.Size*4, module_offset, additional_header)

        self.module_offset = module_offset
        self.huffman_offset = huffman_offset
        return True

    def _parse_variable_mods(self, module_offset):
        # Parse additional tagged modules.
        self.partition_end = 0
        while module_offset < self.structure.Size * 4:
            # There is more module header to process.
            module = MeVariableModule(
                self.data[module_offset:], self.header_type)
            if not module.valid_header:
                break
            module_offset += module.HEADER_SIZE
            if module.header_blank:
                continue

            # print "Debug: found module (%s) size (%d)." % (module.tag,
            # module.size)
            if not module.process():
                return False
            if module.tag == '$MCP':
                # The end of a manifest partition is stored in MCP
                self.partition_end = module.values[0] + module.values[1]
            self.variable_modules.append(module)
            module_offset += module.size
        return True

    def _parse_module_files(self):
        file_offset = self.structure.Size * 4
        # print "Debug: looking for module files at (%08X)." % file_offset
        while True:
            module_file = MeModuleFile(self.data[file_offset:])
            if not module_file.valid_header:
                break
            if module_file.name in module_map:
                # A module file header corresponds to the module header
                self.module_map[module_file.name].file = module_file
            # print "Debug: found module file (%s) size (%d)." %
            # (module_file.name, module_file.size)
            file_offset += module_file.size
        return True

    def process(self):
        self.modules = []
        self.variable_modules = []
        self.module_map = {}

        if self.structure.Tag == '$MN2':
            self.header_type = MeModuleHeader2Type
        elif self.structure.Tag == '$MAN':
            self.header_type = MeModuleHeader1Type
        else:
            # Cannot parse modules...
            return False

        # Parse the module headers (two types of headers, specified by the
        # manifest).
        if not self._parse_mods():
            return False
        if self.header_type == MeModuleHeader2Type:
            if not self._parse_variable_mods(self.module_offset):
                return False
        if not self._parse_module_files():
            return False

        # Parse optional huffman LLUT.
        huffman_offset = self.huffman_offset - self.partition_offset
        huffman_llut = MeLLUT(
            self.data[huffman_offset:], huffman_offset + self.absolute_offset)
        if huffman_llut.valid_header:
            self.huffman_llut = huffman_llut
        #    print "Debug: huffman LLUT start (0x%08X) end (0x%08X)." % (
        #        huffman_llut.offset, huffman_llut.size + huffman_llut.start)
        # print "Debug: LLUT end (%08X) partition end (%08X)." % (
        #    huffman_llut.size + huffman_offset, self.partition_end)
        return True

    def dump(self, parent=""):
        #huffman_end = self.huffman_llut.size + self.huffman_llut.start
        for module in self.modules:
            # if module.compression == COMP_TYPE_HUFFMAN:
                # print "Huffman module data: %r %08X/%08X" % (
                #    module.name, self.huffman_llut.start, self.huffman_llut.size)
            # else:
                #huffman_end = (min(huffman_end, module.structure.Offset))
                # print "Debug: decrementing huffman to %d" % huffman_end
            module.dump(parent)
        if self.huffman_llut is not None:
            self.huffman_llut.dump(
                os.path.join(parent, self.structure.PartitionName))


class CPDEntry(MeObject):

    def __init__(self, data, header_offset):
        self.parse_structure(data[header_offset:], MeCpdEntryType)
        self.valid_header = False
        if self.structure.Offset > len(data):
            # This is invalid, offset (start) out of bounds
            return
        end = self.structure.Offset + self.structure.Size
        if end > len(data):
            # This is invalid, end of data out of bounds
            return
        self.valid_header = True
        self.data = data[self.structure.Offset:end]

    def process(self):
        if not self.valid_header:
            return False
        self.name = self.structure.Name.rstrip('\0')

        # Not sure why the placement of data determines compression type.
        compression = self.structure.Offset >> 24
        if self.name.find('.met') > 0:
            self.compression = COMP_TYPE_NOT_COMPRESSED
        elif compression == 0x02:
            self.compression = COMP_TYPE_HUFFMAN
        elif compression == 0x00:
            self.compression = COMP_TYPE_LZMA
        else:
            self.compression = COMP_TYPE_NOT_COMPRESSED
        return True

    def showinfo(self, ts):
        print "%s%s name= %s offset= 0x%x size= 0x%x (%d bytes) flags= 0x%x" % (
            ts, blue("ME CDP Entry"), purple(self.name),
            self.structure.Offset, self.structure.Size, self.structure.Size,
            self.structure.Flags),
        self.show_compression()

    def dump(self, parent):
        if self.compression == COMP_TYPE_LZMA:
            # There is an odd state to check for that includes an additional
            # \x00\x00\x00 after the initial LZMA header block.
            if self.data[0x0e:0x11] == '\x00\x00\x00':
                self.data = self.data[:0x0e] + self.data[0x11:]
        self.dump_module(parent)


class CPDManifestHeader(MeObject):

    def __init__(self, data, container_offset=0):
        self.valid_header = True
        self.parse_structure(data, MeCpdHeaderType)

        # Save the container offset as LLUT start is an absolute reference
        self.container_offset = container_offset
        self.data = data

        self.partition_name = self.structure.PartitionName.rstrip("\0")
        if not self.partition_name:
            self.partition_name = "(none)"

        self.modules = []

    @property
    def objects(self):
        return self.modules

    def process(self):
        offset = MeCpdHeaderType.size
        for i in xrange(self.structure.NumModules - 1):
            offset += MeCpdEntryType.size
            entry = CPDEntry(self.data, offset)
            if entry.process():
                self.modules.append(entry)
        return True

    def showinfo(self, ts):
        print "%s%s name= %s modules= %d flags= 0x%x" % (
            ts, blue("ME CDP Entry"), purple(self.partition_name),
            self.structure.NumModules, self.structure.Flags)
        for entry in self.modules:
            entry.showinfo("%s  " % ts)

    def dump(self, parent):
        for entry in self.modules:
            entry.dump(parent)


class PartitionEntry(MeObject):
    size = 0x20

    def __init__(self, data, offset):
        self.manifest = None
        self.parse_structure(data[offset:], MeFptEntryType)

        self.has_content = True
        if self.structure.Owner == "\xFF\xFF\xFF\xFF":
            # A blank owner is filled in with 0xFF.
            self.structure.Owner = ''
        if self.structure.Offset == 0xFFFFFFFF:
            # A (blank) offset usually means flags = 0x02.
            self.has_content = False
            return

        # Set the partition data based on an offset and size determined within
        # the partition entry metadata.
        if self.structure.Offset > len(data):
            # This partition is invalid
            self.has_content = False

        partition_end = self.structure.Offset + self.structure.Size
        if partition_end > len(data):
            # This partition is invalid
            self.has_content = False
        self.data = data[self.structure.Offset:partition_end]

    @property
    def objects(self):
        if self.manifest is not None:
            return [self.manifest]
        return []

    def process(self):
        if not self.has_content:
            return True
        if self.data[0:0x04] == '$CPD':
            manifest = CPDManifestHeader(self.data, self.structure.Offset)
        else:
            manifest = MeManifestHeader(self.data, self.structure.Offset)
        if manifest.valid_header:
            if manifest.process():
                self.manifest = manifest
        return True

    def showinfo(self, ts=''):
        print "%s%s name= %s owner= %s offset= 0x%x size= 0x%x (%d bytes) flags= 0x%x" % (
            ts, blue("ME Partition Entry"),
            purple(self.structure.Name), purple(self.structure.Owner),
            self.structure.Offset, self.structure.Size, self.structure.Size, self.structure.Flags)
        if self.manifest is not None:
            self.manifest.showinfo("%s  " % ts)

    def dump(self, parent=""):
        if self.has_content:
            dump_data(os.path.join(parent, "%s.partition" % self.structure.Name),
                self.data)
        if self.manifest is not None:
            self.manifest.dump(os.path.join(parent, self.structure.Name))

class MeContainer(MeObject):

    def __init__(self, data):
        self.partitions = []
        self.data = data

        self.valid_header = False
        if data[0x0:len(ME_HEADER)] == ME_HEADER:
            self.partition_offset = 0x00
            self.valid_header = True
        if data[0x10:0x14] == ME_PARTITION_HEADER:
            self.partition_offset = 0x00
            self.valid_header = True

    @property
    def objects(self):
        return self.partitions

    def process(self):
        self.parse_structure(self.data, MePartitionTable)

        for i in xrange(self.structure.Entries):
            offset = self.partition_offset + 0x30
            offset += i * PartitionEntry.size
            entry = PartitionEntry(self.data, offset)
            if entry.process():
                self.partitions.append(entry)
        return True

    def showinfo(self, ts=''):
        print "%s%s type= 0x%x version= 0x%x size= 0x%x (%d bytes) entires= %d flags= 0x%x" % (
            ts, blue("ME Container"),
            self.structure.Type, self.structure.Version,
            self.structure.Size, self.structure.Size,
            self.structure.Entries, self.structure.Flags)
        for partition in self.partitions:
            partition.showinfo("  %s" % ts)

    def dump(self, parent=""):
        dump_data(os.path.join(parent, "me-container.me"), self.data)
        for partition in self.partitions:
            partition.dump(os.path.join(parent, "partitions"))



import struct
import os

from .base import FirmwareObject, RawObject, BaseObject, AutoRawObject
from .uefi import FirmwareVolume
from .utils import print_error, dump_data, sguid, green, blue


PFS_GUIDS = {
    "FIRMWARE_VOLUMES": "7ec6c2b0-3fe3-42a0-a316-22dd0517c1e8",
    "INTEL_ME":         "7439ed9e-70d3-4b65-9e33-1963a7ad3c37",
    "BIOS_ROMS_1":      "08e56a30-62ed-41c6-9240-b7455ee653d7",
    "BIOS_ROMS_2":      "492261e4-0659-424c-82b6-73274389e7a7"
}


def _discover_volumes(data):
    # Assume a series of firmware volumes
    volumes = []
    fv_offset = 0
    while fv_offset < len(data):
        fv = FirmwareVolume(data[fv_offset:], hex(fv_offset))
        if not fv.valid_header:
            break
        if not fv.process():
            return False
        volumes.append(fv)
        fv_offset += fv.size
    return volumes


class PFSPartitionedSection(FirmwareObject, BaseObject):
    '''A PFSSection with embedded PFSFiles (with additional sections) that
    split the content of the section across multiple chunks. The chunks and the
    encompassing embedded PFSFile are named as 'partitioned' section.

    Search within a PFSFile, similar to handling sets of PFSSections without
    recording information/objects for each section. Instead, strip all data
    and content, save for the defined body. The object then represents the
    single, concatenated body.
    '''

    HEADER_SIZE = 0x48
    DATA_OFFSET = 0x248

    def __init__(self, data):
        self.data = data
        self.size = len(data)
        self.section_objects = []
        self.partitions = 0
        self.section_data = ""

    def process(self):
        # The end removes the PFS trailer.
        body_end = self.size - 0x10
        # This data is the content of a section, with stripped PFS header.
        # The first line will be the UUID.
        self.uuid = self.data[0x0:0x10]
        body_step = 0x10

        # The stepping is equivilent to a PFSSection save for a 0x200-sized
        # set of variables.
        while body_step < body_end:
            # The UUID for partitioned section is useless.
            header = self.data[body_step:body_step + self.HEADER_SIZE]
            if len(header) < self.HEADER_SIZE:
                return False
            self.partitions += 1
            size = struct.unpack("<I", header[0x28:0x28 + 0x04])[0]
            # Advance the seek pointer past the header.
            body_step += self.HEADER_SIZE
            # The section data seeks past an offset of variables.
            data = self.data[body_step + self.DATA_OFFSET:body_step + size]
            self.section_data += data
            sig1_size, trp_size, sig2_size = struct.unpack("<III", header[0x2C:0x2C + 0x0C])
            body_step += size + sig1_size + trp_size + sig2_size

        # Now that section partitions are reconstructed, search for volumes.
        volumes = _discover_volumes(self.section_data)
        if volumes is False:
            return False
        self.section_objects = volumes
        return True

    @property
    def objects(self):
        return self.section_objects

    def showinfo(self, ts='', index=None):
        print "%s%s %s partitions %d size 0x%x (%d bytes)" % (
            ts, blue("Dell PFSPartitionedSection:"), green(sguid(self.uuid)),
            self.partitions, len(self.section_data), len(self.section_data))
        for sub_object in self.section_objects:
            sub_object.showinfo("%s  " % ts)

    def dump(self, parent="", index=None):
        path = os.path.join(parent, "%s" % sguid(self.uuid))
        dump_data("%s.data" % path, self.section_data)

        path = os.path.join(parent, "section-%s" % sguid(self.uuid))
        for sub_object in self.section_objects:
            sub_object.dump(path)
        pass


class PFSSection(FirmwareObject, BaseObject):
    HEADER_SIZE = 0x48

    def __init__(self, data):
        self.data = data
        self.size = -1

        # Store parsed objects (if any)
        self.section_objects = []

    def process(self):
        hdr = self.data[:self.HEADER_SIZE]
        self.uuid = hdr[:0x10]
        self.header = hdr

        # Spec seems to be a consistent 1, what I thought was a timestamp is not.
        # Version is static except for the first section in a PFS
        ##spec, ts, ctype, version, _u1 = struct.unpack("<I4shh4s", hdr[0x10:0x20])
        spec, version_type = struct.unpack("<I4s", hdr[0x10:0x10 + 0x08])
        self.spec = spec
        self.version = ""
        for i in range(4):
            group_offset = 0x18 + (i * 2)
            if version_type[i] == 'A':
                self.version += "%X" % struct.unpack("<h", hdr[group_offset:group_offset + 2])
            elif version_type[i] == 'N':
                self.version += ".%d" % struct.unpack("<h", hdr[group_offset:group_offset + 2])

        # U1, U2 might be flag containers
        _u2, section_size, rsa1_size, pmim_size, rsa2_size = struct.unpack(
            "<8sIIII", hdr[0x20:0x20 + 0x18])

        # This seems to be a set of 8byte CRCs for each chunk (4 total)
        self.crcs = hdr[0x20 + 0x18:self.HEADER_SIZE]
        self.section_data = self.data[
            self.HEADER_SIZE:self.HEADER_SIZE + section_size]

        rsa1_offset = self.HEADER_SIZE + section_size
        self.rsa1 = RawObject(self.data[rsa1_offset:rsa1_offset + rsa1_size])
        pmim_offset = rsa1_offset + rsa1_size
        self.pmim = RawObject(self.data[pmim_offset:pmim_offset + pmim_size])
        rsa2_offset = pmim_offset + pmim_size
        self.rsa2 = RawObject(self.data[rsa2_offset:rsa2_offset + rsa2_size])

        # Unknown 8byte variable
        # _u3 = self.data[64+total_chunk_size:64+total_chunk_size+8]
        self.unknowns = [_u2]

        # Size of header, data, and footer
        total_size = section_size + rsa1_size + pmim_size + rsa2_size
        self.section_size = self.HEADER_SIZE + total_size
        self.data = None

        if self.section_data[:0x08] == "PFS.HDR.":
            # Partitioned ROM
            rom = PFSPartitionedSection(self.section_data)
            if not rom.process():
                return False
            self.section_objects.append(rom)
        elif sguid(self.uuid) == PFS_GUIDS["FIRMWARE_VOLUMES"]:
            volumes = _discover_volumes(self.section_data)
            if volumes is False:
                return False
            self.section_objects += volumes
        else:
            raw = AutoRawObject(self.section_data)
            raw.process()
            self.section_objects.append(raw)

    @property
    def objects(self):
        return self.section_objects + [self.rsa1, self.pmim, self.rsa2]

    def info(self, include_content=False):
        return {
            "_self": self,
            "guid": sguid(self.uuid),
            "type": "PFSSection",
            "content": self.section_data if include_content else "",
            "attrs": {
                "size": self.section_size,
                "crcs": self.crcs.encode("hex"),
                "unknowns": [u.encode("hex") for u in self.unknowns],
                "version": self.version
            },
            "chunks": [self.rsa1, self.pmim, self.rsa2] if include_content else []
        }
        pass

    def build(self, generate_checksum=False, debug=False):
        body = ""
        for sub_object in self.section_objects:
            body += sub_object.build(generate_checksum, debug=debug)
        return self.header + body + \
            self.rsa1.build(generate_checksum) + \
            self.pmim.build(generate_checksum) + \
            self.rsa2.build(generate_checksum)
        pass

    def showinfo(self, ts='', index=None):
        print "%s%s %s spec %d version %s size 0x%x (%d bytes)" % (
            ts, blue("Dell PFSSection:"), green(sguid(self.uuid)),
            self.spec, self.version, self.section_size, self.section_size
        )

        for sub_object in self.section_objects:
            sub_object.showinfo("%s  " % ts)
        pass

    def dump(self, parent="", index=None):
        path = os.path.join(parent, "%s" % sguid(self.uuid))
        dump_data("%s.data" % path, self.section_data)

        # Instead of calling dump on each chunk RawObject, dump with a better
        # name.
        if len(self.rsa1.data) > 0:
            dump_data("%s.rsa1" % path, self.rsa1.data)
        if len(self.pmim.data) > 0:
            dump_data("%s.pmim" % path, self.pmim.data)
        if len(self.rsa2.data) > 0:
            dump_data("%s.rsa2" % path, self.rsa2.data)

        path = os.path.join(parent, "section-%s" % sguid(self.uuid))
        for sub_object in self.section_objects:
            sub_object.dump(path)
        pass


class PFSFile(FirmwareObject):
    PFS_HEADER = "PFS.HDR."
    PFS_FOOTER = "PFS.FTR."

    def __init__(self, data):
        self.sections = []
        self.data = data
        self.valid_header = False
        if self.check_header():
            self.valid_header = True

    def check_header(self):
        if len(self.data) < 32:
            print_error("Data does not contain a header.")
            return False

        header = self.data[:0x10]
        magic, spec, size = struct.unpack("<8sII", header)

        self.spec = spec
        self.size = size

        if magic != self.PFS_HEADER:
            print_error(
                "Data does not contain the header magic (%s)." % self.PFS_HEADER)
            return False

        footer_offset = self.size + 0x10
        footer = self.data[footer_offset:footer_offset + 0x10]
        # Footer size is a repeated body size.
        footer_size, _u2, footer_magic = struct.unpack("<II8s", footer)
        if footer_magic != self.PFS_FOOTER:
            print_error(
                "Data does not container the footer magic (%s)." % self.PFS_FOOTER)
            return False

        return True

    def process(self):
        '''Chunks are assumed to contain a chunk header.'''
        data = self.data[16:-16]
        if not self.valid_header:
            return False

        chunk_num = 0
        offset = 16
        while True:
            section = PFSSection(data)
            section.process()
            self.sections.append(section)

            chunk_num += 1
            offset += section.section_size
            data = data[section.section_size:]

            if len(data) < 64:
                break
        return True

    @property
    def objects(self):
        return self.sections

    def build(self, generate_checksum=False, debug=False):
        body = ""
        for section in self.sections:
            body += section.build(generate_checksum, debug=debug)
        return self.data[:16] + body + self.data[-16:]
        pass

    def showinfo(self, ts='', index=None):
        print "%s%s spec 0x%x size 0x%x (%d bytes)" % (
            ts, blue("DellPFS:"),
            self.spec, self.size, self.size
        )
        for section in self.sections:
            section.showinfo("%s  " % ts)

    def dump(self, parent='', index=None):
        path = os.path.join(parent, "pfsobject.pfs")
        dump_data(path, self.data)

        path = os.path.join(parent, "pfsobject")
        for section in self.sections:
            section.dump(path)



#!/usr/bin/env python

import os
import re
from setuptools import setup, find_packages, Extension, Command


class LintCommand(Command):
    """Run pylint on implementation and test code"""

    description = "Run pylint on implementation and test code"
    user_options = []

    _pylint_options = [
        "--max-line-length 80",
        "--ignore-imports yes"
    ]

    _lint_paths = [
        "uefi_firmware/*.py",
        "uefi_firmware/*/*.py",
        "tests/*.py",
        "scripts/*.py",
        "scripts/contrib/*.py",
    ]

    def initialize_options(self):
        """Set default values for options."""
        pass

    def finalize_options(self):
        """Post-process options."""
        pass

    def run(self):
        """Run the command"""
        os.system("pylint %s %s" % (
            " ".join(self._pylint_options),
            " ".join(self._lint_paths),
        ))

with open('README.rst') as f:
    README = f.read()

with open("uefi_firmware/__init__.py", "r") as f:
    __INIT__ = f.read()

VERSION = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                    __INIT__, re.MULTILINE).group(1)
AUTHOR = re.search(r'^__author__\s*=\s*[\'"]([^\'"]*)[\'"]',
                   __INIT__, re.MULTILINE).group(1)

COMPRESSION_SOURCES = []
COMPRESSION_HEADERS = []
for root, directory, paths in os.walk('uefi_firmware/compression'):
    for path in paths:
        if os.path.splitext(path)[1][1:] == 'h':
            COMPRESSION_HEADERS.append(os.path.join(root, path))
        else:
            COMPRESSION_SOURCES.append(os.path.join(root, path))

setup(
    name='uefi_firmware',
    version=VERSION,
    description='Various data structures and parsing tools for UEFI firmware.',
    long_description=README,
    author=AUTHOR,
    author_email='teddy@prosauce.org',
    url='https://github.com/theopolis/uefi-firmware-parser',
    license='BSD',
    packages=find_packages(exclude=('tests', 'docs')),
    test_suite="tests",
    cmdclass={
        "lint": LintCommand,
    },
    headers=COMPRESSION_HEADERS,
    ext_modules=[
        Extension(
            'uefi_firmware.efi_compressor',
            sources=COMPRESSION_SOURCES,
            include_dirs=[
                os.path.join("uefi_firmware", 'compression', 'Include')
            ],
            depends=COMPRESSION_HEADERS,
        )
    ],
    scripts=[
        'bin/uefi-firmware-parser',
    ],

    classifiers=[
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'Topic :: Security',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    keywords="security uefi firmware parsing bios",
)



'''EFI and UEFI related structures.
This package defines firmware structures for unpacking, decompressing,
extracting, and rebuilding UEFI data.
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import struct

from .base import FirmwareObject, StructuredObject, RawObject, AutoRawObject
from .utils import *
from .guids import get_guid_name
from .structs.uefi_structs import *
from .structs.flash_structs import *

from uefi_firmware import efi_compressor


def _get_file_type(file_type):
    return EFI_FILE_TYPES[file_type] if file_type in EFI_FILE_TYPES else (
        "unknown", "unknown")


def _get_section_type(section_type):
    if section_type in EFI_SECTION_TYPES:
        return EFI_SECTION_TYPES[section_type]
    else:
        return ("unknown", "unknown.bin")


def uefi_name(s):
    '''Return the utf-16le encoded string name for a UEFIFile.'''
    try:
        name = s.decode("utf-16le").split("\0")[0]
        if len(name) == 0:
            return None
        for c in name:
            if ord(c) > 128:
                return None
        return name
    except Exception:
        return None


def compare(data1, data2):
    from hashlib import md5
    md5_1 = md5(data1).hexdigest()
    md5_2 = md5(data2).hexdigest()
    if (md5_1 != md5_2):
        print ("%s != %s" % (red(md5_1), red(md5_2)))
        return False
    return True


def decompress(algorithms, compressed_data):
    '''Attempt to decompress using a set of algorithms.

    Args:
        algorithms (list): A set of decompression methods.
        compressed_data (binary): A compressed data stream.

    Return:
        pair (int, binary): Return the algorithm index, and decompressed stream.
    '''
    for i, algorithm in enumerate(algorithms):
        try:
            data = algorithm(compressed_data, len(compressed_data))
            return (i, data)
        except Exception:
            continue
    return None


def find_volumes(data, process=True):
    '''Search for arbitary firmware volumes within data.

    This is helpful within Raw files and sections.

    Some firmware vendors will implement custom checksums or metadata within
    UEFIFiles. It is often helpful to 'lose' this information and continue to
    discovery volumnes and other structures. In these cases it is ultimately
    helpful if folks/contributors will add support for representing each
    proprietary structure.

    Args:
        process (Optional[bool]): Call process on each discovered volumn.

    Return:
        list: The set of discovered firmware objects.
    '''
    objects = []
    while True:
        volume_index = data.find("_FVH")
        if volume_index < 0:
            break
        volume_index -= (8 + 16 * 2)
        fv = FirmwareVolume(data[volume_index:])
        if not fv.valid_header:
            data = data[16 * 3:]
            continue
        if volume_index > 0:
            objects.append(RawObject(data[:volume_index]))
        if process:
            fv.process()
        objects.append(fv)
        data = data[volume_index + fv.size:]
    if len(data) > 0:
        objects.append(RawObject(data))
    return objects


class FirmwareVariableStore(FirmwareObject, StructuredObject):

    '''An firmware-related variable storage structure (think NVRAM).'''
    variables = []

    @property
    def objects(self):
        return self.variables


class FirmwareVariable(FirmwareObject, StructuredObject):

    '''A firmware-related variable, found in a variable store.'''
    subsections = []

    @property
    def objects(self):
        return self.subsections


class NVARVariable(FirmwareVariable):

    @classmethod
    def valid_nvar(cls, data):
        if data[:4] != "NVAR":
            return False
        return True

    def _get_name(self, data, is_ascii=False):
        tail = "\x00" if is_ascii else "\x00\x00"
        size = data.find(tail)
        if not is_ascii:
            name = uefi_name(data[:size])
        else:
            name = data[:size]
        return (name, size + len(tail))

    def __init__(self, data):
        self.subsections = []
        self.name = None
        self.guid = None
        self.data = data

    def process(self):
        if not NVARVariable.valid_nvar(self.data):
            return False
        self.parse_structure(self.data, NVARVariableHeaderType)
        self.size = self.structure.TotalSize
        self.attrs = {"attrs": self.structure.Attributes}

        # Now with structure parsed, set bounds on the data
        self.data = self.data[:self.size]
        offset = self.structure_size
        if bit_set(self.structure.Attributes, NVRAM_ATTRIBUTES["GUID"]):
            self.guid = self.data[offset:offset + 16]
            offset += 16
        else:
            # Increment data by 1!
            offset += 1

        if bit_set(self.structure.Attributes, NVRAM_ATTRIBUTES["DATA"]):
            self.data_offset = self.size - self.structure_size
            # self.subsections.append(RawObject(self.data[offset:]))
            return True

        # Parse variable name.
        var_name, var_name_size = self._get_name(
            self.data[offset:],
            bit_set(self.structure.Attributes, NVRAM_ATTRIBUTES["DESC_ASCII"])
        )
        if var_name is not None:
            self.name = var_name
            offset += var_name_size

        # The variable's meta-data is the value.
        self.data_offset = self.size - self.structure_size

        # if self.data_offset+1 < self.size:
        #    self.subsections.append(RawObject(self.data[offset:]))
        return True

    def build(self, generate_checksum=False, debug=False):
        header = self.structure_data
        data = ""
        for section in self.subsections:
            data += section.build(generate_checksum, debug)
        # Metadata includes optional guid/name.
        meta_data = self.data[self.structure_size:self.data_offset]
        return header + meta_data + data

    def dump(self, parent, index=0):
        path = os.path.join(parent, "variable%d.nvar" % index)
        dump_data(path, self.data)
        for i, section in enumerate(self.subsections):
            section.dump(os.path.join(parent, "variable%d-data" % index), i)

    def showinfo(self, ts="", index=0):
        '''Potential for A LOT of variables.'''
        if self.guid is not None and self.name is not None:
            print ("%s %s %s %s" % (
                blue("%sVariable:" % ts),
                green(sguid(self.guid)),
                purple(self.name),
                "attrs= %s" % self.attrs["attrs"]
            ))


class NVARVariableStore(FirmwareVariableStore):
    '''NVAR has no header, only a series of variable headers.'''

    def __init__(self, data):
        self.variables = []
        self.valid_header = False
        if not NVARVariable.valid_nvar(data):
            return
        self.data = data
        self.valid_header = True

    def process(self):
        if not self.valid_header:
            return False

        var_offset = self.data
        total_size = 0
        while len(var_offset) > 4:
            nvar = NVARVariable(var_offset)
            if not nvar.process():
                break
            total_size += nvar.size
            self.variables.append(nvar)
            var_offset = var_offset[nvar.size:]

        # Scope data to just the parsed variables
        self.data = self.data[:total_size]
        self.attrs = {"variables": len(self.variables)}
        return True

    def build(self, generate_checksum=False, debug=False):
        data = ""
        for variable in self.variables:
            data += variable.build(generate_checksum, debug)
        return data

    def dump(self, parent, index=0):
        if not self.valid_header:
            return
        path = os.path.join(parent, "nvar.vars")
        dump_data(path, self.data)
        for i, variable in enumerate(self.variables):
            variable.dump(parent, i)
        pass

    def showinfo(self, ts="", index=0):
        if not self.valid_header:
            return
        print ("%s %s" % (
            blue("%sNVAR Variable Store:" % ts),
            "variables: %d" % self.attrs["variables"]
        ))
        for i, variable in enumerate(self.variables):
            variable.showinfo("%s  " % ts, i)


class EfiSection(FirmwareObject):
    subsections = []

    @property
    def objects(self):
        return self.subsections

    def process_subsections(self):
        self.subsections = []

        if self.data is None:
            return False

        subsection_offset = 0
        status = True
        while subsection_offset < len(self.data):
            if subsection_offset % 4:
                subsection_offset += 4 - (subsection_offset % 4)
            if subsection_offset >= len(self.data):
                break

            try:
                subsection = FirmwareFileSystemSection(
                    self.data[subsection_offset:],
                    self.guid
                )
            except struct.error:
                return False
            if subsection.size == 0:
                break
            status = subsection.process() and status
            self.subsections.append(subsection)

            subsection_offset += subsection.size
        return status

    def build(self, generate_checksum=False, debug=False):
        raise Exception("Cannot build from unknown section type!")

    def process(self):
        pass

    def showinfo(self, ts='', index=-1):
        pass

    def dump(self, parent="", index=0):
        for i, subsection in enumerate(self.subsections):
            subsection.dump(parent, i)

    def _build_subsections(self, generate_checksum=False):
        data = ""
        for i, section in enumerate(self.subsections):
            subsection_size, subsection_data = section.build(generate_checksum)
            data += subsection_data
            if (i + 1 < len(self.subsections)):
                # Nibble-align inter-section subsections
                data += "\x00" * \
                    (((subsection_size + 3) & (~3)) - subsection_size)

        # Pad the pre-compression data
        trailling_bytes = len(self.data) - len(data)
        if trailling_bytes > 0:
            data += '\x00' * trailling_bytes
        return data


class CompressedSection(EfiSection):
    name = None

    ATTR_NOT_COMPRESSED = 0x00
    ATTR_STANDARD_COMPRESSION = 0x01
    ATTR_CUSTOMIZED_COMPRESSION = 0x02

    def __init__(self, data, guid):
        self.guid = guid
        self.data = None
        self.parsed_objects = []

        # http://dox.ipxe.org/PiFirmwareFile_8h_source.html
        self.decompressed_size, self.type = struct.unpack("<Ic", data[:5])
        self.type = ord(self.type)
        # A special compression type to determine (EFI/Tiano if type= 0x01).
        self.subtype = 0

        # Advance the byte pointer through the header
        self.compressed_data = data[5:]
        self.attrs = {
            "decompressed_size": self.decompressed_size, "type": self.type}

    def process(self):
        def bf_decompress(data):
            return decompress([
                efi_compressor.LzmaDecompress,
                efi_compressor.TianoDecompress,
                efi_compressor.EfiDecompress,
            ], data)

        if self.type == 0x00:
            '''No compression.'''
            self.data = self.compressed_data

        results = None
        if self.type == 0x01:
            # Tiano or Efi compression, unfortunately these are identified by
            # the same byte
            results = decompress([
                efi_compressor.EfiDecompress,
                efi_compressor.TianoDecompress,
            ], self.compressed_data)
        if self.type == 0x02:
            results = bf_decompress(self.compressed_data)
            if results is None and len(self.compressed_data) > 4:
                # The type=2 is not spec-defined, may have an additional int
                # (Intel).
                results = bf_decompress(self.compressed_data[4:])

        if self.type > 0x00:
            if results is not None:
                self.subtype = results[0] + 1
                self.data = results[1]
            else:
                print_error(
                    "Cannot EFI decompress GUID (%s), type= (%d), decompressed_size= (%d)" % (
                        sguid(self.guid),
                        self.type,
                        self.decompressed_size
                    )
                )
                raw = AutoRawObject(self.compressed_data)
                raw.process()
                self.subsections.append(raw)

        if self.data is None:
            '''No data was uncompressed.'''
            return True

        status = self.process_subsections()
        return status
        pass

    def build(self, generate_checksum=False, debug=False):
        data = self._build_subsections()

        if self.type == 0x01:
            if self.subtype == 0x01:
                data = str(efi_compressor.EfiCompress(data, len(data)))
            elif self.subtype == 0x02:
                data = str(efi_compressor.TianoCompress(data, len(data)))
        elif self.type == 0x02:
            data = str(efi_compressor.LzmaCompress(data, len(data)))
        elif self.type == 0x00:
            pass

        header = struct.pack("<Ic", self.decompressed_size, chr(self.type))
        return header + data
        pass

    def showinfo(self, ts):
        if self.name is not None:
            print ("%s %s" % (blue("%sCompressed Name:" % ts), purple(self.name)))
        for i, _object in enumerate(self.subsections):
            _object.showinfo(ts, i)


class VersionSection(EfiSection):

    def __init__(self, data):
        self.data = data
        self.build_number = struct.unpack("<16s", self.data[:16])


class FreeformGuidSection(EfiSection):
    '''A firmware file section type (free-form GUID)

    struct { UCHAR GUID[16]; }
    '''

    name = None

    def __init__(self, data):
        self.guid = struct.unpack("<16s", data[:16])[0]
        self.data = data[16:]

    def process(self):
        if sguid(self.guid) == FIRMWARE_FREEFORM_GUIDS["CHAR_GUID"]:
            self.guid_header = self.data[:12]
            self.name = uefi_name(self.data[12:])
        return True

    def build(self, generate_checksum=False, debug=False):
        # print "Building FreeformGUID: %s" % green(sguid(self.guid))

        header = struct.pack("<16s", self.guid)
        return header + self.data

    def showinfo(self, ts='', index=-1):
        # print "%sGUID: %s" % (ts, green(sguid(self.guid)))
        if self.name is not None:
            print ("%sGUID Description: %s" % (ts, purple(self.name)))


class GuidDefinedSection(EfiSection):
    '''A firmware file section type (GUID-defined)

    struct { UCHAR GUID[16]; short offset; short attrs; }
    '''

    ATTR_PROCESSING_REQUIRED = 0x01
    ATTR_AUTH_STATUS_VALID = 0x02

    def __init__(self, data):
        self.guid, self.offset, self.attr_mask = struct.unpack(
            "<16sHH", data[:20])

        # A guid-defined section includes an offset
        self.preamble = data[20:self.offset]
        self.data = data[self.offset:]
        self.attrs = {"attrs": self.attr_mask}
        self.subsections = []

    @property
    def objects(self):
        return self.subsections

    def process(self):
        def parse_volume():
            fv = FirmwareVolume(self.data)
            if fv.valid_header:
                fv.process()
                self.subsections = [fv]
                return True
            return False

        def decompress_guid(alg):
            # Try to decompress the body of the section.
            results = decompress([alg], self.preamble + self.data)
            if results is None:
                # Attempt to recover by skipping the preamble.
                results = decompress([alg], self.data)
                if results is None:
                    return False
            self.subtype = results[0] + 1
            self.data = results[1]
            return self.process_subsections()

        status = True
        if sguid(self.guid) == FIRMWARE_GUIDED_GUIDS["LZMA_COMPRESSED"]:
            status = decompress_guid(efi_compressor.LzmaDecompress)
        if sguid(self.guid) == FIRMWARE_GUIDED_GUIDS["TIANO_COMPRESSED"]:
            status = decompress_guid(efi_compressor.TianoDecompress)
        # Todo: check for processing required attribute
        elif sguid(self.guid) == FIRMWARE_GUIDED_GUIDS["STATIC_GUID"]:
            # Todo: verify this (FirmwareFile hack)
            self.data = self.preamble[-4:] + self.data
            status = self.process_subsections()
            if len(self.subsections) == 0:
                # There were no subsections parsed, treat as a firmware volume
                status = parse_volume()
                if not status:
                    raw = AutoRawObject(self.data)
                    raw.process()
                    self.subsections.append(raw)
            pass
        elif sguid(self.guid) == FIRMWARE_GUIDED_GUIDS["FIRMWARE_VOLUME"]:
            status = parse_volume()
        else:
            # Undefined GUIDed-Section GUID, treat as a FV, don't require
            # status
            parse_volume()
        return status
        pass

    def build(self, generate_checksum=False, debug=False):
        data = self._build_subsections(generate_checksum)

        if sguid(self.guid) == FIRMWARE_GUIDED_GUIDS["LZMA_COMPRESSED"]:
            data = str(efi_compressor.LzmaCompress(data, len(data)))
            pass

        header = struct.pack(
            "<16sHH", self.guid, self.offset, self.attrs["attrs"])
        return header + self.preamble + data

    def showinfo(self, ts='', index=0):
        auth_status = "ATTR_UNKNOWN"
        if self.attrs["attrs"] == self.ATTR_AUTH_STATUS_VALID:
            auth_status = "AUTH_VALID"
        if self.attrs["attrs"] == self.ATTR_PROCESSING_REQUIRED:
            auth_status = "PROCESSING_REQUIRED"
        print ("%s%s %s offset= 0x%x attrs= 0x%x (%s)" % (
            ts, blue("Guid-Defined:"), green(sguid(self.guid)),
            self.offset, self.attrs["attrs"], purple(auth_status)
        ))
        if len(self.subsections) > 0:
            for i, section in enumerate(self.subsections):
                section.showinfo("%s  " % ts, index=i)

    def dump(self, parent="", generate_checksum=False, debug=False):
        for i, subsection in enumerate(self.subsections):
            subsection.dump(parent, i)
        dump_data(os.path.join(parent, "guided.preamble"), self.preamble)
        dump_data(os.path.join(parent, "guided.certs"), self.preamble[172:])
        pass

    pass


class FirmwareFileSystemSection(EfiSection):
    '''A firmware file section

    struct { UINT8 Size[3]; EFI_SECTION_TYPE Type; } EFI_COMMON_SECTION_HEADER;
    '''

    parsed_object = None
    '''For object sections, keep track of each.'''

    def __init__(self, data, guid):
        self.guid = guid
        header = data[:0x4]

        self.valid_header = True
        try:
            self.size, self.type = struct.unpack("<3sB", header)
            self.size = struct.unpack("<I", self.size + "\x00")[0]
        except Exception:
            print_error("%s: invalid FFS Section header, expected size 4, got (%d)." % (
                red("Error"),
                len(header)
            ))
            self.valid_header = False
            return

        self._data = data[:self.size]
        self.data = data[0x4:self.size]
        self.name = None

    @property
    def objects(self):
        return [self.parsed_object]

    def regen(self, data):
        # Transitional method, should be adopted by other objects.
        self._data = data
        self.data = data[0x4:]

    def process(self):
        self.parsed_object = None

        if self.type == 0x01:  # compression
            compressed_section = CompressedSection(self.data, self.guid)
            self.parsed_object = compressed_section

        elif self.type == 0x02:  # GUID-defined
            guid_defined = GuidDefinedSection(self.data)
            self.parsed_object = guid_defined

        elif self.type == 0x14:  # version string
            self.name = uefi_name(self.data)

        elif self.type == 0x15:  # user interface name
            self.name = uefi_name(self.data)

        elif self.type == 0x17:  # firmware-volume
            fv = FirmwareVolume(self.data, sguid(self.guid))
            if not fv.valid_header:
                # Could be a FFSv3 section (Kairos sample)
                fv = FirmwareVolume(self.data[4:], sguid(self.guid))
            if fv.valid_header:
                self.parsed_object = fv

        elif self.type == 0x18:  # freeform GUID
            freeform_guid = FreeformGuidSection(self.data)
            self.parsed_object = freeform_guid

        elif self.type == 0x19:  # raw
            if self.data[:10] == "123456789A":
                # HP adds a strange header to nested FVs.
                fv = FirmwareVolume(self.data[12:], sguid(self.guid))
                self.parsed_object = fv
            else:
                # For a raw section, we can cheat and assign the parsed object
                # as the AutoRawObject's managed object
                raw = AutoRawObject(self.data)
                raw.process()
                if raw.object is not None:
                    self.parsed_object = raw.object

        self.attrs = {"type": self.type, "size": self.size}
        self.attrs["type_name"] = _get_section_type(self.type)[0]

        if self.parsed_object is None:
            return True
        status = self.parsed_object.process()
        return status

    def build(self, generate_checksum=False, debug=False):
        data = ""
        # Add section data (either raw, or a partitioned section)
        if self.parsed_object is not None:
            data = self.parsed_object.build(generate_checksum)
        else:
            data = self.data

        # Pad the data and check for potential overflows.
        size = self.size
        trailling_bytes = (self.size - 4) - len(data)
        if trailling_bytes > 0:
            data += '\x00' * trailling_bytes
        if trailling_bytes < 0:
            size = self.size - trailling_bytes
            # raise Exception("FileSystemSection GUID %s has overflown %d bytes." % (
            #    sguid(self.guid), trailling_bytes*-1))
            pass

        string_size = struct.pack("<I", size)
        header = struct.pack("<3sB", string_size[:3], self.type)
        return size, header + data
        pass

    def showinfo(self, ts='', index=-1):
        print ("%s type 0x%02x, size 0x%x (%d bytes) (%s section)" % (
            blue("%sSection %d:" % (ts, index)),
            self.type, self.size, self.size,
            _get_section_type(self.type)[0]
        ))
        if self.type == 0x15 and self.name is not None:
            print ("%sName: %s" % (ts, purple(self.name)))

        if self.parsed_object is not None:
            '''If this is a specific object, show that object's info.'''
            self.parsed_object.showinfo(ts + '  ')

    def dump(self, parent="", index=0):
        self.path = os.path.join(
            parent, "section%d.%s" % (index, _get_section_type(self.type)[1]))
        dump_data(self.path, self.data)

        if self.parsed_object is not None:
            self.parsed_object.dump(os.path.join(parent, "section%d" % index))


class FirmwareFile(FirmwareObject):
    '''A firmware file is contained within a firmware file system and is comprised of firmware file
    sections.

    struct {
        UCHAR: FileNameGUID[16]
        UINT16: Checksum (header/file)
        UINT8: Filetype
        UINT8: Attributes
        UINT8: Size[3]
        UINT8: State
    };
    '''
    _HEADER_SIZE = 0x18  # 24 byte header, always

    def __init__(self, data):
        header = data[:self._HEADER_SIZE]

        try:
            self.guid, self.checksum, self.type, self.attributes, \
                self.size, self.state = struct.unpack("<16sHBB3sB", header)
            self.size = struct.unpack("<I", self.size + "\x00")[0]
        except Exception as e:
            print_error("Error: invalid FirmwareFile header.")
            raise e

        self.attrs = {
            "size": self.size,
            "type": self.type,
            "attributes": self.attributes,
            "state": self.state ^ 0xFF
        }
        self.attrs["type_name"] = _get_file_type(self.type)[0]

        # The size includes the header bytes.
        self._data = data[:self.size]
        self.data = data[self._HEADER_SIZE:self.size]
        self.raw_blobs = []
        self.sections = []

    @property
    def objects(self):
        invalid_types = [bytes, str, unicode]
        valid_blobs = [
            b for b in self.raw_blobs if type(b) not in invalid_types]
        return self.sections + valid_blobs

    def regen(self, data):
        # Transitional method, should be adopted by other objects.
        self.__init__(data)

    def process(self):
        '''Parse the file and file sections if appropriate.'''
        if self.type == 0xf0:  # ffs padding
            return True

        status = True
        if sguid(self.guid) == FIRMWARE_VOLUME_GUIDS["NVRAM_NVAR"]:
            var_store = NVARVariableStore(self.data)
            if not var_store.valid_header:
                raw = AutoRawObject(self.data)
                raw.process()
                self.raw_blobs.append(raw)
            else:
                status = var_store.process()
                self.raw_blobs.append(var_store)
            return status

        if self.type == 0x01:  # raw file
            status = self._find_objects()
            return status

        if self.type == 0x00:  # unknown
            raw = AutoRawObject(self.data)
            raw.process()
            self.raw_blobs.append(raw)
            return True

        section_data = self.data
        self.sections = []
        while len(section_data) >= 4:
            file_section = FirmwareFileSystemSection(section_data, self.guid)
            if not file_section.valid_header:
                return False
            if file_section.size <= 0:
                # This is not expected, something bad happened while parsing.
                print_error("Error: file section size <= 0 (%d)." %
                            file_section.size)
                return False

            status = file_section.process() and status
            self.sections.append(file_section)

            section_data = section_data[(file_section.size + 3) & (~3):]
        return status

    def _find_objects(self):
        '''Helper function for wacky raw type usages.'''
        has_object = False
        status = True

        # It may be a firmware volume (Lenovo or HP).
        fv = FirmwareVolume(self.data, sguid(self.guid))
        if fv.valid_header:
            has_object = True
            status = fv.process() and status
            self.raw_blobs.append(fv)
        elif self.data[0x10:0x10 + 4] == FLASH_HEADER:
            # Lenovo may also bundle a flash descriptor as raw content.
            from .flash import FlashDescriptor
            flash = FlashDescriptor(self.data)
            if flash.valid_header:
                has_object = True
                status = flash.process() and status
                self.raw_blobs.append(flash)

        # If everything is normal (according to the FV/FF spec).
        if not has_object:
            # There may be arbitrary firmware structures (Lenovo)
            objects = find_volumes(self.data)
            self.raw_blobs += objects
            return True
        return status

    def build(self, generate_checksum=False, debug=False):
        data = ""
        for i, section in enumerate(self.sections):
            section_size, section_data = section.build(generate_checksum)
            data += section_data
            if (i + 1 < len(self.sections)):
                # Nibble-align inter-file sections
                data += "\x00" * (((section_size + 3) & (~3)) - section_size)

        for blob in self.raw_blobs:
            if isinstance(blob, FirmwareObject):
                data += blob.build(generate_checksum)
            elif isinstance(blob, AutoRawObject) or isinstance(blob, RawObject):
                data += blob.data
            else:
                data += blob

        # Maining to support ffs-padding
        if len(self.raw_blobs) == 0 and len(self.sections) == 0:
            data = self.data

        if generate_checksum:
            pass

        size = self.size
        trailling_bytes = size - (len(data) + 24)
        if trailling_bytes < 0:
            print ("%s adding %s-bytes to GUID: %s" % (
                red("Warning"),
                red(trailling_bytes * -1),
                red(sguid(self.guid))
            ))
            size += (trailling_bytes * -1)

        string_size = struct.pack("<I", size)
        header = struct.pack(
            "<16sHBB3sB",
            self.guid, self.checksum, self.type, self.attributes, string_size[
                :3], self.state
        )
        return size, header + data

    def showinfo(self, ts='', index="N/A"):
        guid_name = get_guid_name(self.guid)
        if guid_name is None:
            guid_display = "%s" % green(sguid(self.guid))
        else:
            guid_display = "%s (%s)" % (
                green(sguid(self.guid)), purple(guid_name))
        print ("%s %s type 0x%02x, attr 0x%02x, state 0x%02x, size 0x%x (%d bytes), (%s)" % (
            blue("%sFile %s:" % (ts, index)),
            guid_display,
            self.type,
            self.attributes,
            self.state ^ 0xFF,
            self.size,
            self.size,
            _get_file_type(self.type)[0]
        ))

        for i, blob in enumerate(self.raw_blobs):
            if type(blob) not in [str, bytes]:
                blob.showinfo(ts + "  ", index=i)
            else:
                self._guessinfo(ts + "  ", blob, index=i)

        if self.sections is None:
            # padding file, skip for now
            return

        for i, section in enumerate(self.sections):
            section.showinfo(ts + "  ", index=i)

    def _guessinfo(self, ts, data, index="N/A"):
        if data[:4] == "\x01\x00\x00\x00" and data[
                20:24] == "\x01\x00\x00\x00":
            print ("%s Might contain CPU microcodes" % (blue("%sBlob %d:" % (ts, index))))

    def dump(self, parent=""):
        parent = os.path.join(parent, "file-%s" % sguid(self.guid))

        dump_data(os.path.join(parent, "file.obj"), self._data)
        if self.raw_blobs is not None:
            for i, blob in enumerate(self.raw_blobs):
                blob.dump(parent, index=i)

        if self.sections is not None:
            for i, section in enumerate(self.sections):
                section.dump(parent, index=i)


class FirmwareFileSystem(FirmwareObject):
    '''A potential UEFI firmware filesystem (FFS) data stream.

    The FFS is a specific GUID within the FirmwareVolume.
    '''

    def __init__(self, data):
        self.files = []
        self._data = data

        # Overflow data is non-file data within the filesystem
        self.overflow_data = ""

    @property
    def objects(self):
        return self.files or []

    def process(self):
        '''Search for a 24-byte header that does not contain all 0xFF.'''

        data = self._data
        status = True
        while len(data) >= 24 and data[:24] != ("\xff" * 24):
            firmware_file = FirmwareFile(data)

            if firmware_file.size < 24:
                # This is a problem, the file was corrupted.
                break

            status = firmware_file.process() and status
            self.files.append(firmware_file)
            data = data[(firmware_file.size + 7) & (~7):]

        if len(data) > 0:
            # There is overflow data
            self.overflow_data = data
        return status

    def build(self, generate_checksum=False, debug=False):

        # Generate the file system data as an unstructed set of file data.
        data = ""
        for firmware_file in self.files:
            file_size, file_data = firmware_file.build(generate_checksum)
            data += file_data
            data += "\xFF" * (((file_size + 7) & (~7)) - file_size)

        data += self.overflow_data

        if len(data) != len(self._data):
            print ("ffs size mismatch old=%d new=%d %d" % (
                len(self._data),
                len(data),
                len(self._data) - len(data)
            ))

        return data
        pass

    def showinfo(self, ts=''):
        for i, firmware_file in enumerate(self.files):
            firmware_file.showinfo(ts + ' ', index=i)

    def dump(self, parent=""):
        dump_data(os.path.join(parent, "filesystem.ffs"), self._data)

        for _file in self.files:
            _file.dump(parent)


class FirmwareVolume(FirmwareObject):
    '''Describes the features and layout of the firmware volume.

    struct EFI_FIRMWARE_VOLUME_HEADER {
        UINT8: Zeros[16]
        UCHAR: FileSystemGUID[16]
        UINT64: Length
        UINT32: Signature (_FVH)
        UINT8: Attribute mask
        UINT16: Header Length
        UINT16: Checksum
        UINT8: Reserved[3]
        UINT8: Revision
        [<BlockMap>]+, <BlockMap(0,0)>
    };

    The block map is a set of blocks followed by a zeroed block indicating the
    end of the map set.

    struct BLOCK_MAP {
        UINT32: Block count
        UINT32: Block size
    };

    '''

    _HEADER_SIZE = 0x38

    name = None
    '''string: An optional name or offset of the firmware volume.'''

    block_map = None
    '''list: An empty block set.'''

    firmware_filesystems = []
    '''list: Set of FirmwareFileSystems discovered in volume.'''

    raw_objects = []
    '''list: Set of RawObjects discovered in volume.'''

    def __init__(self, data, name="volume"):
        self.firmware_filesystems = []
        self.raw_objects = []
        self.name = name
        self.valid_header = False

        try:
            header = data[:self._HEADER_SIZE]
            self.rsvd, self.guid, self.size, self.magic, self.attributes, \
                self.hdrlen, self.checksum, self.rsvd2, \
                self.revision = struct.unpack("<16s16sQ4sIHH3sB", header)
        except Exception as e:
            # print "Error: cannot parse FV header (%s)." % str(e)
            return

        if sguid(self.guid) not in FIRMWARE_VOLUME_GUIDS.values():
            # print "Error: invalid FV guid (%s)." % sguid(self.guid)
            return

        self.blocks = []
        self.block_map = ""

        try:
            data = data[:self.size]

            self._data = data
            self.data = data[self.hdrlen:]
            self.block_map = data[self._HEADER_SIZE:self.hdrlen]
        except Exception as e:
            print_error("Error invalid FV header data (%s)." % str(e))
            return

        self.valid_header = True
        pass

    @property
    def objects(self):
        return self.firmware_filesystems or []

    def process(self):
        if self.block_map is None:
            return False

        block_data = self.block_map
        while len(block_data) > 0:
            block = block_data[:8]

            block_size, block_length = struct.unpack("<II", block)
            if (block_size, block_length) == (0, 0):
                '''The block map ends with a (0, 0) block.'''
                break

            self.blocks.append((block_size, block_length))
            block_data = block_data[8:]

        if len(self.blocks) == 0:
            '''No block in the volume? This is a problem.'''
            return False

        data = self.data
        self.firmware_filesystems = []
        self.raw_objects = []
        status = True

        ffs_guids = [
            FIRMWARE_VOLUME_GUIDS["FFS1"],
            FIRMWARE_VOLUME_GUIDS["FFS2"],
            FIRMWARE_VOLUME_GUIDS["FFS3"],
        ]
        for block in self.blocks:
            if sguid(self.guid) in ffs_guids:
                firmware_filesystem = FirmwareFileSystem(
                    data[:block[0] * block[1]])
                status = firmware_filesystem.process() and status
                self.firmware_filesystems.append(firmware_filesystem)
            elif sguid(self.guid) == FIRMWARE_VOLUME_GUIDS["NVRAM_EVSA"]:
                # If this is an NVRAM volume, there are no FFS/FFs.
                self.raw_objects.append(
                    NVARVariableStore(data[:block[0] * block[1]]))
            else:
                self.raw_objects.append(data[:block[0] * block[1]])
            data = data[block[0] * block[1]:]
        return status

    def build(self, generate_checksum=False, debug=False):
        # Generate blocks from FirmwareFileSystems
        data = ""
        for filesystem in self.firmware_filesystems:
            # print "Building filesystem"
            data += filesystem.build(generate_checksum)

        # Generate block map from original block map (assume no size change)
        block_map = ""
        for block in self.blocks:
            block_map += struct.pack("<II", block[0], block[1])
        # Add a trailing-NULL to the block map
        block_map += "\x00" * 8

        if generate_checksum:
            pass

        # Assume no size change
        header = struct.pack(
            "<16s16sQ4sIHH3sB",
            self.rsvd, self.guid, self.size,
            self.magic, self.attributes, self.hdrlen, self.checksum, self.rsvd2, self.revision
        )
        return header + block_map + data
        pass

    def showinfo(self, ts='', index=None):
        if not self.valid_header or len(self.data) == 0:
            return

        print ("%s %s attr 0x%08x, rev %d, cksum 0x%x, size 0x%x (%d bytes)" % (
            blue("%sFirmware Volume:" % (ts)),
            green(sguid(self.guid)),
            self.attributes,
            self.revision,
            self.checksum,
            self.size,
            self.size
        ))
        print (blue("%s  Firmware Volume Blocks: " % (ts)), end="")
        for block_size, block_length in self.blocks:
            print ("(%d, 0x%x)" % (block_size, block_length), end="")
        print ("")

        for _ffs in self.firmware_filesystems:
            _ffs.showinfo(ts + " ")
        for raw in self.raw_objects:
            print ("%s%s NVRAM" % ("%s  " % ts, blue("Raw section:")))

    def dump(self, parent="", index=None):
        if len(self.data) == 0:
            return

        path = os.path.join(parent, "volume-%s.fv" % self.name)
        dump_data(path, self._data)

        for _ffs in self.firmware_filesystems:
            _ffs.dump(os.path.join(parent, "volume-%s" % self.name))


class FirmwareCapsule(FirmwareObject):
    '''EFI Capsule Header.

    struct EFI_CAPSULE_HEADER {
        UCHAR:  CapsuleGUID[16]
        UINT32: HeaderSize
        UINT32: Flags
        UINT32: CapsureImageSize
        UINT32: SequenceNumber
        UCHAR:  InstanceGUID[16]
        UINT32: OffsetToSplitInformation
        UINT32: OffsetToCapsuleBody
        UINT32: OffsetToOemDefinedHeader
        UINT32: OffsetToAuthorInformation
        UINT32: OffsetToRevisionInformation
        UINT32: OffsetToShortDescription
        UINT32: OffsetToLongDescription
        UINT32: OffsetToApplicableDevices
    }
    '''

    capsule_body = None
    '''binary: Data string of the capsule content.'''

    def __init__(self, data, name="Capsule"):
        self.name = name
        self.valid_header = True
        self.data = None

        self.capsule_guid = data[:16]
        self.guid = "\x00" * 16
        if sguid(self.capsule_guid) not in FIRMWARE_CAPSULE_GUIDS:
            self.valid_header = False
            return

        try:
            self.parse_capsule_header(data[16:])
        except:
            self.valid_header = False
            return

        # Header sections
        self.header_sections = []

        # Set data (original, and body content)
        self._data = data
        self.data = data[self.header_size:]

        pass

    def parse_capsule_header(self, data):
        if sguid(self.capsule_guid) == FIRMWARE_CAPSULE_GUIDS[0]:
            # EFICapsule
            self.size, self.flags, self.image_size, self.seq_num = struct.unpack(
                "<IIII",
                data[:4 * 4]
            )
            self.guid = data[16:32]
            split_info, capsule_body, oem_header, author_info, revision_info, \
                short_desc, long_desc, compatibility = struct.unpack(
                    "<" + "I" * 8,
                    data[32:32 + 4 * 8]
                )

            # Store offsets
            self.offsets = {
                # This offset can be relative to the base of the capsule or end
                # of the header.
                "capsule_body": capsule_body,
                "split_info": split_info,
                "oem_header": oem_header,
                "author_info": author_info,
                "revision_info": revision_info,
                "short_desc": short_desc,
                "long_desc": long_desc,
                "compatibility": compatibility
            }
        elif sguid(self.capsule_guid) == FIRMWARE_CAPSULE_GUIDS[1]:
            # EFI2Capsule
            self.size, self.flags, self.image_size = struct.unpack(
                "<III", data[:4 * 3])
            fv_image, oem_header = struct.unpack("<HH", data[12:12 + 4])
            self.offsets = {
                "capsule_body": fv_image,
                "oem_header": oem_header,
                "author_info": 0
            }
        elif sguid(self.capsule_guid) == FIRMWARE_CAPSULE_GUIDS[2]:
            # UEFI Capsule
            self.size, self.flags, self.image_size = struct.unpack(
                "<III", data[:4 * 3])
            self.offsets = {
                "capsule_body": self.size,
                "oem_header": 0,
                "author_info": 0
            }

        self.header_size = self.size
        pass

    def parse_sections(self, header):
        # Parse the various pieces within the capsule header, before body
        pass

    @property
    def objects(self):
        return [self.capsule_body]

    def process(self):
        # Copy the EOH to capsule into a preamble
        self.preamble = self.data[:self.offsets["capsule_body"]]
        self.parse_sections(None)

        fv = FirmwareVolume(self.data[self.offsets["capsule_body"]:])
        if not fv.valid_header:
            # The body could be an offset from the end of the header (Intel
            # does this).
            fv = FirmwareVolume(
                self.data[self.offsets["capsule_body"] - self.header_size:])
            if not fv.valid_header:
                return False

        if not fv.process():
            # Todo: test code coverage
            # return False
            pass
        self.capsule_body = fv
        return True

    def build(self, generate_checksum=False, debug=False):
        if self.capsule_body is not None:
            body = self.capsule_body.build(generate_checksum, debug=debug)
        else:
            body = self.data[self.offsets["capsule_body"]:]

        # Assume no size change
        return self._data[:self.header_size] + self.preamble + body
        pass

    def showinfo(self, ts='', index=None):
        if not self.valid_header or len(self.data) == 0:
            return

        print ("%s %s flags 0x%08x, size 0x%x (%d bytes)" % (
            blue("%sFirmware Capsule:" % (ts)),
            "%s/%s" % (green(sguid(self.capsule_guid)),
                       green(sguid(self.guid))),
            self.flags, self.size, self.size
        ))
        print ("%s  Details: size= 0x%x (%d bytes) body= 0x0%x, oem= 0x0%x, author= 0x0%x" % (
            ts, self.image_size, self.image_size,
            self.offsets["capsule_body"], self.offsets[
                "oem_header"], self.offsets["author_info"]
        ))
        # print self.offsets

        if self.capsule_body is not None:
            self.capsule_body.showinfo(ts)
        pass

    def dump(self, parent="", index=None):
        if len(self.data) == 0:
            return

        path = os.path.join(parent, "capsule-%s.cap" % self.name)
        dump_data(path, self._data)

        if self.capsule_body is not None:
            self.capsule_body.dump(
                os.path.join(parent, "capsule-%s" % self.name))
        else:
            # Write the raw image data from the capsule.
            path = os.path.join(parent, "capsule-%s.image" % self.name)
            offset = self.offsets["capsule_body"]
            dump_data(path, self.data[offset:offset + self.image_size])



"""
These are misc functions/classes to implement several type checkers.
The TypeTester may be useful if parsing a large number of UEFI-related binaries.
"""

from __future__ import absolute_import

import re

from ..uefi import FirmwareVolume, FirmwareCapsule
from ..pfs import PFSFile
from ..flash import FlashDescriptor
from ..me import MeContainer, MeManifestHeader


class TypeTester(object):
    parser = None
    static = "MZ"

    def match(self, data):
        if data[:self.size] == self.static:
            return True
        return False

    @property
    def size(self):
        return len(self.static)

    @property
    def name(self):
        return self.__class__.__name__.replace("Tester", "")


class UEFIFirmwareVolumeTester(TypeTester):
    parser = FirmwareVolume

    def match(self, data):
        fv = FirmwareVolume(data)
        return fv.valid_header


class FlashDescriptorTester(TypeTester):
    parser = FlashDescriptor

    def match(self, data):
        fd = FlashDescriptor(data)
        return fd.valid_header


class EFICapsuleTester(TypeTester):
    static = "".join(
        "BD 86 66 3B 76 0D 30 40 B7 0E B5 51 9E 2F C5 A0".split(" ")).decode('hex')
    parser = FirmwareCapsule


class UEFICapsuleTester(TypeTester):
    static = "".join(
        "B9 82 91 53 B5 AB 91 43 B6 9A E3 A9 43 F7 2F CC".split(" ")).decode('hex')
    parser = FirmwareCapsule


class IntelMEPartitionManifestTester(TypeTester):
    static = "".join("04 00 00 00 A1 00 00 00".split(" ")).decode('hex')
    parser = MeManifestHeader


class IntelMETester(TypeTester):
    parser = MeContainer

    def match(self, data):
        me = MeContainer(data)
        return me.valid_header


class DellPFSTester(TypeTester):
    static = "PFS.HDR"
    parser = PFSFile


class DellUpdateBinaryTester(TypeTester):
    hdr_pattern = re.compile(
        r'.{4}\xAA\xEE\xAA\x76\x1B\xEC\xBB\x20\xF1\xE6\x51.{1}\x78\x9C')
    static = "\x00" * 100

    def match(self, data):
        hdr_match = self.hdr_pattern.search(data)
        if hdr_match is None:
            return False
        return True

TESTERS = [
    UEFIFirmwareVolumeTester,
    FlashDescriptorTester,
    UEFICapsuleTester,
    EFICapsuleTester,
    IntelMEPartitionManifestTester,
    IntelMETester,
    DellPFSTester,
    DellUpdateBinaryTester,
]



# -*- coding: utf-8 -*-

import os
import struct

from .base import FirmwareObject, BaseObject, StructuredObject
from .me import MeContainer
from .utils import *
from .structs.flash_structs import *


class RegionSection(StructuredObject):
    size = 20

    def __init__(self, data):
        self.parse_structure(data, FlashRegionSectionType)


class MasterSection(StructuredObject):
    size = 12

    def __init__(self, data):
        self.parse_structure(data, FlashMasterSectionType)


class DescriptorMap(StructuredObject):
    size = 16

    def __init__(self, data):
        self.parse_structure(data, FlashDescriptorMapType)


class FlashRegion(FirmwareObject, BaseObject):

    def __init__(self, data, region_name, region_details):
        self.sections = []
        self.data = data
        self.attrs = region_details
        self.name = region_name

    @property
    def objects(self):
        return self.sections

    def process(self):
        from .uefi import FirmwareVolume

        if self.name == "bios":
            data = self.data
            while True:
                volume_index = search_firmware_volumes(data, limit=1)
                if len(volume_index) == 0:
                    break
                fv = FirmwareVolume(data[volume_index[0] - 40:])
                if fv.valid_header:
                    self.sections.append(fv)
                    data = data[volume_index[0] - 40 + fv.size:]
                else:
                    data = data[volume_index[0] + 8:]
        if self.name == "me":
            data = self.data
            me = MeContainer(data)
            if me.valid_header:
                self.sections.append(me)
        for section in self.sections:
            section.process()
        return True

    def showinfo(self, ts='', index=None):
        print "%s%s type= %s, size= 0x%x (%d bytes) details[ %s ]" % (
            ts, blue("Flash Region"), green(self.name),
            len(self.data), len(self.data),
            ", ".join(["%s: %s" % (k, v) for k, v in self.attrs.iteritems()])
        )
        for section in self.sections:
            section.showinfo(ts="%s  " % ts)
        pass

    def dump(self, parent=""):
        dump_data(os.path.join(parent, "region-%s.fd" % self.name), self.data)

        parent = os.path.join(parent, "region-%s" % self.name)
        for section in self.sections:
            section.dump(parent)
        pass
    pass


class FlashDescriptor(FirmwareObject):

    def __init__(self, data):
        self.valid_header = False
        if len(data) < 20:
            return

        self.padding, self.header = struct.unpack("<16s4s", data[:16 + 4])
        if self.header != FLASH_HEADER:
            return

        self.valid_header = True
        self.regions = []
        self.data = data

    @property
    def objects(self):
        return self.regions

    def process(self):
        def _region_size(base, limit):
            if limit:
                return (limit + 1 - base) * 0x1000
            return 0

        def _region_offset(base):
            return base * 0x1000

        self.map = DescriptorMap(self.data[20:20 + DescriptorMap.size])
        region_offset = (self.map.structure.RegionBase * 0x10)
        self.region = RegionSection(
            self.data[region_offset:region_offset + RegionSection.size])
        master_offset = (self.map.structure.MasterBase * 0x10)
        self.master = MasterSection(
            self.data[master_offset:master_offset + MasterSection.size])

        bios_base = self.region.structure.BiosBase
        bios_limit = self.region.structure.BiosLimit
        bios_size = _region_offset(
            bios_base) + _region_size(bios_base, bios_limit)
        bios = self.data[_region_offset(bios_base): bios_size]

        bios_region = FlashRegion(bios, "bios", {
            "base": bios_base,
            "limit": bios_limit,
            "id": self.master.structure.BiosId,
            "read": self.master.structure.BiosRead,
            "write": self.master.structure.BiosWrite
        })
        bios_region.process()
        self.regions.append(bios_region)

        me_base = self.region.structure.MeBase
        me_limit = self.region.structure.MeLimit
        me_size = _region_offset(me_base) + _region_size(me_base, me_limit)
        me = self.data[_region_offset(me_base): me_size]

        me_region = FlashRegion(me, "me", {
            "base": me_base,
            "limit": me_limit,
            "id": self.master.structure.MeId,
            "read": self.master.structure.MeRead,
            "write": self.master.structure.MeWrite
        })
        me_region.process()
        self.regions.append(me_region)

        gbe_base = self.region.structure.GbeBase
        gbe_limit = self.region.structure.GbeLimit
        gbe_size = _region_offset(gbe_base) + _region_size(gbe_base, gbe_limit)
        gbe = self.data[_region_offset(gbe_base): gbe_size]

        gbe_region = FlashRegion(gbe, "gbe", {
            "base": gbe_base,
            "limit": gbe_limit,
            "id": self.master.structure.GbeId,
            "read": self.master.structure.GbeRead,
            "write": self.master.structure.GbeWrite
        })
        gbe_region.process()
        self.regions.append(gbe_region)
        
        pdr_base = self.region.structure.PdrBase
        pdr_limit = self.region.structure.PdrLimit
        pdr_size = _region_offset(pdr_base) + _region_size(pdr_base, pdr_limit)
        pdr = self.data[_region_offset(pdr_base): pdr_size]

        pdr_region = FlashRegion(pdr, "pdr", {
            "base": pdr_base,
            "limit": pdr_limit,
        })
        pdr_region.process()
        self.regions.append(pdr_region)
        return True

    def showinfo(self, ts='', index=None):
        print (("%s%s chips %d, regions %d, masters %d, PCH straps %d, "
                "PROC straps %d, ICC entries %d") % (
            ts, blue("Flash Descriptor (Intel PCH)"),
            self.map.structure.NumberOfFlashChips,
            self.map.structure.NumberOfRegions,
            self.map.structure.NumberOfMasters,
            self.map.structure.NumberOfPchStraps,
            self.map.structure.NumberOfProcStraps,
            self.map.structure.NumberOfIccTableEntries))
        for region in self.regions:
            region.showinfo(ts="%s  " % ts)

    def dump(self, parent, index=None):
        dump_data(os.path.join(parent, "flash.fd"), self.data)

        parent = os.path.join(parent, "regions")
        for region in self.regions:
            region.dump(parent)

    pass
